"""Pytest translation of the unittest-based `test_issue_139.py`.

Uses mth5-provided fixtures where available to be xdist-safe and fast.

This test writes a TF z-file (zrr) from an in-memory TF object generated
from a synthetic MTH5 file, reads it back, and asserts numeric equality
of primary arrays.
"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from mt_metadata.transfer_functions.core import TF

from aurora.test_utils.synthetic.processing_helpers import tf_obj_from_synthetic_data


@pytest.fixture
def tf_obj_from_mth5(fresh_test12rr_mth5: Path):
    """Create a TF object from the provided fresh `test12rr` MTH5 file.

    Uses the `fresh_test12rr_mth5` fixture (created by the mth5 `conftest.py`).
    """
    return tf_obj_from_synthetic_data(fresh_test12rr_mth5)


def write_and_read_zrr(tf_obj: TF, zrr_path: Path) -> TF:
    """Write `tf_obj` to `zrr_path` as a zrr file and read it back as TF."""
    tf_obj.run_metadata.channels["hy"].measurement_azimuth = 90
    tf_obj.run_metadata.channels["ey"].measurement_azimuth = 90
    # write expects a filename; TF.write will create the zrr
    tf_obj.write(fn=str(zrr_path), file_type="zrr")

    tf_z = TF()
    tf_z.read(str(zrr_path))
    return tf_z


def _register_cleanup(cleanup_test_files, p: Path):
    try:
        cleanup_test_files(p)
    except Exception:
        # Best-effort: if the helper isn't available, ignore
        pass


def test_write_and_read_zrr(
    tf_obj_from_mth5,
    make_worker_safe_path,
    cleanup_test_files,
    tmp_path: Path,
    subtests,
):
    """Round-trip a TF through a `.zrr` write/read and validate arrays.

    This test uses `make_worker_safe_path` to generate a worker-unique
    filename so it is safe to run under `pytest-xdist`.
    """

    # Create a worker-safe path in the tmp directory
    zrr_path = make_worker_safe_path("synthetic_test1.zrr", tmp_path)

    # register cleanup so sessions don't leak files
    _register_cleanup(cleanup_test_files, zrr_path)

    # Write and read back
    tf_z = write_and_read_zrr(tf_obj_from_mth5, zrr_path)

    # Use subtests to make multiple assertions clearer in pytest output
    with subtests.test("transfer_function_data"):
        assert (
            np.isclose(
                tf_z.transfer_function.data,
                tf_obj_from_mth5.transfer_function.data,
                atol=1e-4,
            )
        ).all()

    with subtests.test("period_arrays"):
        assert np.allclose(tf_z.period, tf_obj_from_mth5.period)

    with subtests.test("shape_checks"):
        assert (
            tf_z.transfer_function.data.shape
            == tf_obj_from_mth5.transfer_function.data.shape
        )



"""
This module contains methods that are used to define datasets to build from FDSN servers.

These datasets are in turn used for testing.

"""

from obspy import UTCDateTime

from aurora.sandbox.io_helpers.fdsn_dataset import FDSNDataset


def make_pkdsao_test_00_config(minitest=False) -> FDSNDataset:
    """
    Return a description of a 2h PKD SAO 40Hz dataset from NCEDC.

    Parameters
    ----------
    minitest: bool
        Used for debugging, when set to True, it will just get a minute of data

    Returns
    -------
    test_data_set: aurora.sandbox.io_helpers.fdsn_dataset.FDSNDataset

    """
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "pkd_sao_test_00"
    test_data_set.network = "BK"
    test_data_set.station = "PKD,SAO"
    test_data_set.starttime = UTCDateTime("2004-09-28T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2004-09-28T02:00:00.000000Z")
    if minitest:
        test_data_set.endtime = UTCDateTime("2004-09-28T00:01:00")  # 1 min
    test_data_set.channel_codes = "BQ2,BQ3,BT1,BT2,BT3"
    test_data_set.description = "2h of PKD,SAO data for 2004-09-28 0000-0200 UTC"
    test_data_set.components_list = ["ex", "ey", "hx", "hy", "hz"]
    return test_data_set


def make_cas04_nvr08_test_00_config() -> FDSNDataset:
    """
    Return a description of a CAS04,NVR08 dataset from IRIS.

    Returns
    -------

    """
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "cas_nvr_test_00"
    test_data_set.network = "ZU"
    test_data_set.station = "CAS04,NVR08"
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2020-06-02T18:41:43.000000Z")
    # test_data_set.endtime = UTCDateTime("2020-07-13T21:46:12.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2020-06-04T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2020-06-05T00:00:00.000000Z")  # minitest
    # test_data_set.endtime = UTCDateTime("2020-06-24T15:55:46.000000Z")

    # test_data_set.starttime = UTCDateTime("2004-09-28T00:00:00.000000Z")
    # test_data_set.endtime = UTCDateTime("2004-09-28T01:59:59.975000Z")
    # test_data_set.endtime = UTCDateTime("2004-09-28T00:01:59.999000Z") #small test
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_iak34_test_00_config() -> FDSNDataset:
    """Return a description of a IAK34 dataset from IRIS."""
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "iak34_test_00"
    test_data_set.network = "EM"
    test_data_set.station = "IAK34"
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2013-04-25T20:10:08.000000Z")
    # test_data_set.endtime = UTCDateTime("2013-05-13T21:18:53.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2013-04-26T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-05-12T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-04-27T00:00:00.000000Z")
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset IAK34"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_iak34_test_01_config() -> FDSNDataset:
    """Return a description of a IAK34 dataset from IRIS."""
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "iak34_test_01_long_ss"
    test_data_set.network = "EM"
    test_data_set.station = "IAK34"
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2013-04-25T20:10:08.000000Z")
    # test_data_set.endtime = UTCDateTime("2013-05-13T21:18:53.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2013-04-26T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-05-11T00:00:00.000000Z")
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset IAK34"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_iak34_test_02_config() -> FDSNDataset:
    """Return a description of a IAK34 dataset from IRIS."""
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "iak34_test_02_long_rr"
    test_data_set.network = "EM"
    test_data_set.station = "IAK34,NEK33"
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2013-04-25T20:10:08.000000Z")
    # test_data_set.endtime = UTCDateTime("2013-05-13T21:18:53.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2013-04-26T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-05-10T00:00:00.000000Z")
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset IAK34"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_iak34_test_03_config() -> FDSNDataset:
    """Return a description of a IAK34 dataset from IRIS."""
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "iak34_test_03_long_rr"
    test_data_set.network = "EM"
    test_data_set.station = "IAK34,NEK33"
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2013-04-25T20:10:08.000000Z")
    # test_data_set.endtime = UTCDateTime("2013-05-13T21:18:53.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2013-05-15T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-05-26T00:00:00.000000Z")
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset IAK34"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_iak34_test_04_config() -> FDSNDataset:
    """Return a description of a IAK34 dataset from IRIS."""
    test_data_set = FDSNDataset()
    test_data_set.dataset_id = "iak34_test_04_rr"
    test_data_set.network = "EM"
    test_data_set.station = "IAK34,NEN34"  # NEK33
    # <ORIGINAL>
    # test_data_set.starttime = UTCDateTime("2013-04-25T20:10:08.000000Z")
    # test_data_set.endtime = UTCDateTime("2013-05-13T21:18:53.000000Z")
    # </ORIGINAL>
    test_data_set.starttime = UTCDateTime("2013-04-28T00:00:00.000000Z")
    test_data_set.endtime = UTCDateTime("2013-04-29T00:00:00.000000Z")
    test_data_set.channel_codes = None
    test_data_set.description = "earthscope example dataset IAK34"
    test_data_set.components_list = ["hx", "hy", "ex", "ey"]
    return test_data_set


def make_test_configs() -> dict:
    """Make all the test dataset configs and put them in a dict"""
    test_data_set_configs = {}

    # pkd_sao_test_00 Remote Reference
    test_data_set = make_pkdsao_test_00_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set

    # cas_nvr_test_00 Remote Reference
    test_data_set = make_cas04_nvr08_test_00_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set

    # IAK34SS
    test_data_set = make_iak34_test_00_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set
    test_data_set = make_iak34_test_01_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set
    test_data_set = make_iak34_test_02_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set
    test_data_set = make_iak34_test_03_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set
    test_data_set = make_iak34_test_04_config()
    test_data_set_configs[test_data_set.dataset_id] = test_data_set

    return test_data_set_configs


TEST_DATA_SET_CONFIGS = make_test_configs()



import pathlib
import unittest

from aurora.sandbox.triage_metadata import triage_run_id
from mth5.mth5 import MTH5
from mth5.utils.helpers import initialize_mth5


class TestTriageMetadata(unittest.TestCase):
    """ """

    @classmethod
    def setUpClass(cls):
        cls.fn = pathlib.Path("tmp.h5")
        cls.station_id = "MT1"
        cls.expected_run_id = "001"
        cls.erroneous_run_id = "123"
        m = initialize_mth5(cls.fn)
        m.add_station(cls.station_id)
        m.add_run(cls.station_id, cls.erroneous_run_id)
        m.close_mth5()

    def test_triage_run_id(self):
        m = MTH5()
        m.open_mth5(self.fn)
        run_obj = m.get_run(self.station_id, self.erroneous_run_id)
        assert run_obj.metadata.id == self.erroneous_run_id
        triage_run_id(self.expected_run_id, run_obj)
        assert run_obj.metadata.id == self.expected_run_id
        return

    @classmethod
    def tearDownClass(cls) -> None:
        cls.fn.unlink()


# def main():
#     unittest.main()
#
#
# if __name__ == "__main__":
#     main()



"""
TODO: Deprecate -- This now basically duplicates a test in MTH5 (issue #191)

Tests setting of start time as per aurora issue #188
"""

import logging

import pandas as pd
import pytest
from loguru import logger
from mth5.data.station_config import make_station_03
from mth5.helpers import close_open_files
from mth5.processing import RunSummary


@pytest.fixture(autouse=True)
def setup_logging():
    """Disable noisy matplotlib loggers."""
    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True


@pytest.fixture(scope="module")
def run_summary_test3(worker_safe_test3_h5):
    """Create a RunSummary from test3.h5 MTH5 file."""
    close_open_files()
    mth5_paths = [worker_safe_test3_h5]
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    return run_summary


def test_start_times_correct(run_summary_test3, subtests):
    """Test that start times in run summary match station configuration."""
    station_03 = make_station_03()

    for run in station_03.runs:
        with subtests.test(run=run.run_metadata.id):
            summary_row = run_summary_test3.df[
                run_summary_test3.df.run == run.run_metadata.id
            ].iloc[0]
            logger.info(summary_row.start)
            logger.info(run.run_metadata.time_period.start)
            assert summary_row.start == pd.Timestamp(
                str(run.run_metadata.time_period.start)
            )



"""Pytest suite for Parkfield dataset processing and calibration tests.

This module tests:
- Calibration and spectral analysis for Parkfield data
- Single-station transfer function processing with various clock_zero configurations
- Remote-reference transfer function processing
- Channel summary conversion helpers
- Comparison with EMTF reference results

Tests are organized into classes and use fixtures from conftest.py for efficient
resource sharing and pytest-xdist compatibility.
"""

import tempfile
from pathlib import Path

import numpy as np
import pytest
from loguru import logger
from mth5.mth5 import MTH5

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.sandbox.mth5_channel_summary_helpers import channel_summary_to_make_mth5
from aurora.test_utils.parkfield.calibration_helpers import parkfield_sanity_check
from aurora.time_series.windowing_scheme import WindowingScheme
from aurora.transfer_function.compare import CompareTF


# ============================================================================
# Helper Functions
# ============================================================================


def _compare_transfer_functions_with_emtf(
    processed_tf,
    z_file_path,
    parkfield_paths,
    tmp_path,
    subtests,
    file_type="zss",
    plot_name="comparison.png",
    z_std_limit=6.8,
):
    """Helper function to compare aurora results with EMTF reference.

    Args:
        processed_tf: Processed transfer function object
        z_file_path: Path to write z-file
        parkfield_paths: Dictionary with paths to test data
        tmp_path: Temporary directory path
        subtests: pytest subtests fixture
        file_type: Type of z-file ("zss" or "zrr")
        plot_name: Name for comparison plot
        z_std_limit: Threshold for impedance standard deviation
    """
    tf_cls = processed_tf
    tf_cls.write(fn=z_file_path, file_type=file_type)

    if not z_file_path.exists():
        pytest.skip("Z-file not generated - data access issue")

    # Compare with archived EMTF results
    auxiliary_z_file = parkfield_paths["emtf_results"].joinpath("PKD_272_00.zrr")
    if not auxiliary_z_file.exists():
        pytest.skip("EMTF reference file not available")

    compare = CompareTF(z_file_path, auxiliary_z_file)

    # Create comparison plot
    output_png = tmp_path / plot_name
    logger.info(f"Comparison plot path: {output_png}")
    compare.plot_two_transfer_functions(save_plot_path=output_png)

    assert output_png.exists()

    # Compare transfer functions numerically
    result = compare.compare_transfer_functions()

    # Check impedance if present
    z_ratio = (0.8, 1.2)
    if result["impedance_ratio"] is not None:
        for ii in range(2):
            for jj in range(2):
                if ii != jj:
                    key = f"Z_{ii}{jj}"
                    with subtests.test(
                        msg=f"Checking impedance magnitude ratio for {key}"
                    ):
                        assert (
                            z_ratio[0] < result["impedance_ratio"][key] < z_ratio[1]
                        ), f"{key} impedance magnitudes differ significantly. Median ratio: {result['impedance_ratio'][key]:.3f}"

                    with subtests.test(msg=f"Checking impedance std for {key}"):
                        assert (
                            result["impedance_std"][key] < z_std_limit
                        ), f"{key} impedance magnitudes have high standard deviation: {result['impedance_std'][key]:.3f}"

    # Check tipper if present
    t_ratio = (0.8, 1.2)
    t_std_limit = 0.5
    if result["tipper_ratio"] is not None:
        for ii in range(2):
            for jj in range(2):
                if ii != jj:
                    key = f"T_{ii}{jj}"
                    with subtests.test(
                        msg=f"Checking tipper magnitude ratio for {key}"
                    ):
                        assert (
                            t_ratio[0] < result["tipper_ratio"][key] < t_ratio[1]
                        ), f"{key} tipper magnitudes differ significantly. Median ratio: {result['tipper_ratio'][key]:.3f}"

                    with subtests.test(msg=f"Checking tipper std for {key}"):
                        assert (
                            result["tipper_std"][key] < t_std_limit
                        ), f"{key} tipper magnitudes have high standard deviation: {result['tipper_std'][key]:.3f}"


# ============================================================================
# Calibration Tests
# ============================================================================


class TestParkfieldCalibration:
    """Test calibration and spectral analysis for Parkfield data."""

    @pytest.fixture
    def windowing_scheme(self, parkfield_run_ts_pkd):
        """Create windowing scheme for spectral analysis.

        Use the actual data length for the window. Should be exactly 2 hours
        (288000 samples at 40 Hz).
        """
        actual_data_length = parkfield_run_ts_pkd.dataset.time.shape[0]
        return WindowingScheme(
            taper_family="hamming",
            num_samples_window=actual_data_length,
            num_samples_overlap=0,
            sample_rate=parkfield_run_ts_pkd.sample_rate,
        )

    @pytest.fixture
    def fft_obj(self, parkfield_run_ts_pkd, windowing_scheme):
        """Compute FFT of Parkfield run data."""
        windowed_obj = windowing_scheme.apply_sliding_window(
            parkfield_run_ts_pkd.dataset, dt=1.0 / parkfield_run_ts_pkd.sample_rate
        )
        tapered_obj = windowing_scheme.apply_taper(windowed_obj)
        return windowing_scheme.apply_fft(tapered_obj)

    def test_windowing_scheme_properties(self, windowing_scheme, parkfield_run_ts_pkd):
        """Test windowing scheme is configured correctly."""
        assert windowing_scheme.taper_family == "hamming"
        assert windowing_scheme.num_samples_window == 288000
        assert windowing_scheme.num_samples_overlap == 0
        assert windowing_scheme.sample_rate == 40.0

    def test_fft_has_expected_channels(self, fft_obj):
        """Test FFT object contains all expected channels."""
        expected_channels = ["ex", "ey", "hx", "hy", "hz"]
        channel_keys = list(fft_obj.data_vars.keys())
        for channel in expected_channels:
            assert channel in channel_keys

    def test_fft_has_frequency_coordinate(self, fft_obj):
        """Test FFT object has frequency coordinate."""
        assert "frequency" in fft_obj.coords
        frequencies = fft_obj.frequency.data
        assert len(frequencies) > 0
        assert frequencies[0] >= 0  # Should start at DC or near-DC

    def test_calibration_sanity_check(
        self, fft_obj, parkfield_run_pkd, parkfield_paths, disable_matplotlib_logging
    ):
        """Test calibration produces valid results."""
        # This should not raise exceptions
        parkfield_sanity_check(
            fft_obj,
            parkfield_run_pkd,
            figures_path=parkfield_paths["aurora_results"],
            show_response_curves=False,
            show_spectra=False,
            include_decimation=False,
        )

    def test_calibrated_spectra_are_finite(self, fft_obj, parkfield_run_pkd):
        """Test that calibrated spectra contain no NaN or Inf values."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Run calibration
            parkfield_sanity_check(
                fft_obj,
                parkfield_run_pkd,
                figures_path=Path(tmpdir),
                show_response_curves=False,
                show_spectra=False,
                include_decimation=False,
            )

        # If we get here without exceptions, calibration succeeded
        # The parkfield_sanity_check function already validates the calibration


# ============================================================================
# Single-Station Processing Tests
# ============================================================================


class TestParkfieldSingleStation:
    """Test single-station transfer function processing."""

    @pytest.fixture
    def z_file_path(self, tmp_path, worker_id, make_worker_safe_path):
        """Generate worker-safe path for z-file output."""
        return make_worker_safe_path("pkd_ss.zss", tmp_path)

    @pytest.fixture(scope="class")
    def config_ss(self, parkfield_kernel_dataset_ss):
        """Create single-station processing config."""
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(
            parkfield_kernel_dataset_ss,
            estimator={"engine": "RME"},
            output_channels=["ex", "ey"],
        )
        return config

    @pytest.fixture(scope="class")
    def processed_tf_ss(self, parkfield_kernel_dataset_ss, config_ss):
        """Process single-station transfer function once and reuse.

        This fixture is class-scoped to avoid reprocessing for each test.
        Processing takes ~2 minutes, so reusing saves significant time.
        """
        tf_cls = process_mth5(
            config_ss,
            parkfield_kernel_dataset_ss,
            units="MT",
            show_plot=False,
        )
        return tf_cls

    def test_single_station_default_processing(
        self,
        processed_tf_ss,
        z_file_path,
        disable_matplotlib_logging,
    ):
        """Test single-station processing with default settings."""
        # Use pre-computed transfer function
        tf_cls = processed_tf_ss

        # Write z-file for verification
        tf_cls.write(fn=z_file_path, file_type="zss")

        assert tf_cls is not None
        assert z_file_path.exists()

        # Verify transfer function has expected properties
        assert hasattr(tf_cls, "station")
        assert hasattr(tf_cls, "transfer_function")

    def test_single_station_clock_zero_configurations(
        self, parkfield_kernel_dataset_ss, subtests, disable_matplotlib_logging
    ):
        """Test single-station processing with different clock_zero settings."""
        clock_zero_configs = [
            {"type": None, "value": None},
            {"type": "data start", "value": None},
            {"type": "user specified", "value": "2004-09-28 00:00:10+00:00"},
        ]

        for clock_config in clock_zero_configs:
            with subtests.test(clock_zero_type=clock_config["type"]):
                cc = ConfigCreator()
                config = cc.create_from_kernel_dataset(
                    parkfield_kernel_dataset_ss,
                    estimator={"engine": "RME"},
                    output_channels=["ex", "ey"],
                )

                # Apply clock_zero configuration
                if clock_config["type"] is not None:
                    for dec_lvl_cfg in config.decimations:
                        dec_lvl_cfg.stft.window.clock_zero_type = clock_config["type"]
                        if clock_config["type"] == "user specified":
                            dec_lvl_cfg.stft.window.clock_zero = clock_config["value"]

                try:
                    tf_cls = process_mth5(
                        config,
                        parkfield_kernel_dataset_ss,
                        units="MT",
                        show_plot=False,
                    )
                    # Processing may skip if insufficient data after clock_zero truncation
                    # Just verify it doesn't crash
                except Exception as e:
                    pytest.fail(f"Processing failed: {e}")

    def test_single_station_emtfxml_export(
        self,
        processed_tf_ss,
        parkfield_paths,
        disable_matplotlib_logging,
    ):
        """Test exporting transfer function to EMTF XML format.

        Currently skipped due to bug in mt_metadata EMTFXML writer (data.py:385):
        IndexError when tipper error arrays have size 0. The writer tries to
        access array[index] even when array has shape (0,).
        """
        tf_cls = processed_tf_ss

        output_xml = parkfield_paths["aurora_results"].joinpath("emtfxml_test_ss.xml")
        output_xml.parent.mkdir(parents=True, exist_ok=True)

        # Use 'xml' as file_type (emtfxml format is accessed via xml)
        tf_cls.write(fn=output_xml, file_type="xml")
        assert output_xml.exists()

    # @pytest.mark.skip(
    #     reason=(
    #         "Archived results seem to have a different coordinate system or a minus "
    #         "sign floating around.  The apparent resistivities are close but the phases "
    #         "are not.  Skipping test for now until a more robust test is created."
    #     )
    # )
    def test_single_station_comparison_with_emtf(
        self,
        processed_tf_ss,
        parkfield_paths,
        tmp_path,
        disable_matplotlib_logging,
        subtests,
    ):
        """Test comparison of aurora results with EMTF reference."""
        z_file_path = tmp_path / "pkd_ss_comparison.zss"
        logger.info(f"Z-file path for comparison: {z_file_path}")
        _compare_transfer_functions_with_emtf(
            processed_tf_ss,
            z_file_path,
            parkfield_paths,
            tmp_path,
            subtests,
            file_type="zss",
            plot_name="SS_processing_comparison.png",
            z_std_limit=6.8,
        )


# ============================================================================
# Remote Reference Processing Tests
# ============================================================================


class TestParkfieldRemoteReference:
    """Test remote-reference transfer function processing."""

    @pytest.fixture(scope="class")
    def config_rr(self, parkfield_kernel_dataset_rr):
        """Create remote-reference processing config."""
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(
            parkfield_kernel_dataset_rr,
            output_channels=["ex", "ey"],
        )
        return config

    @pytest.fixture(scope="class")
    def processed_tf_rr(self, parkfield_kernel_dataset_rr, config_rr):
        """Process remote-reference transfer function once and reuse.

        This fixture is class-scoped to avoid reprocessing for each test.
        """
        tf_cls = process_mth5(
            config_rr,
            parkfield_kernel_dataset_rr,
            units="MT",
            show_plot=False,
            return_collection=False,
        )
        return tf_cls

    def test_remote_reference_processing(
        self,
        processed_tf_rr,
        tmp_path,
        disable_matplotlib_logging,
    ):
        """Test remote-reference processing with SAO as reference."""
        z_file_path = tmp_path / "pkd_rr.zrr"
        tf_cls = processed_tf_rr
        tf_cls.write(fn=z_file_path, file_type="zrr")

        assert tf_cls is not None
        assert z_file_path.exists()

    def test_rr_comparison_with_emtf(
        self,
        processed_tf_rr,
        parkfield_paths,
        tmp_path,
        disable_matplotlib_logging,
        subtests,
    ):
        """Test RR comparison of aurora results with EMTF reference."""
        z_file_path = tmp_path / "pkd_rr_comparison.zrr"
        _compare_transfer_functions_with_emtf(
            processed_tf_rr,
            z_file_path,
            parkfield_paths,
            tmp_path,
            subtests,
            file_type="zrr",
            plot_name="RR_processing_comparison.png",
            z_std_limit=6.8,
        )


# ============================================================================
# Helper Function Tests
# ============================================================================


class TestParkfieldHelpers:
    """Test helper functions used in Parkfield processing."""

    def test_channel_summary_to_make_mth5(
        self, parkfield_h5_path, disable_matplotlib_logging
    ):
        """Test channel_summary_to_make_mth5 helper function."""
        mth5_obj = MTH5(file_version="0.1.0")
        mth5_obj.open_mth5(parkfield_h5_path, mode="r")
        df = mth5_obj.channel_summary.to_dataframe()

        make_mth5_df = channel_summary_to_make_mth5(df, network="NCEDC")

        assert make_mth5_df is not None
        assert len(make_mth5_df) > 0
        assert "station" in make_mth5_df.columns

        mth5_obj.close_mth5()


# ============================================================================
# Data Integrity Tests
# ============================================================================


class TestParkfieldDataIntegrity:
    """Test data integrity and expected properties of Parkfield dataset."""

    def test_mth5_file_exists(self, parkfield_h5_path):
        """Test that Parkfield MTH5 file exists."""
        assert parkfield_h5_path.exists()
        assert parkfield_h5_path.suffix == ".h5"

    def test_pkd_station_exists(self, parkfield_mth5):
        """Test PKD station exists in MTH5 file."""
        station_list = parkfield_mth5.stations_group.groups_list
        assert "PKD" in station_list

    def test_sao_station_exists(self, parkfield_mth5):
        """Test SAO station exists in MTH5 file."""
        station_list = parkfield_mth5.stations_group.groups_list
        assert "SAO" in station_list

    def test_pkd_run_001_exists(self, parkfield_mth5):
        """Test run 001 exists for PKD station."""
        station = parkfield_mth5.get_station("PKD")
        run_list = station.groups_list
        assert "001" in run_list

    def test_pkd_channels(self, parkfield_run_pkd):
        """Test PKD run has expected channels."""
        expected_channels = ["ex", "ey", "hx", "hy", "hz"]
        channels = parkfield_run_pkd.groups_list

        for channel in expected_channels:
            assert channel in channels

    def test_pkd_sample_rate(self, parkfield_run_ts_pkd):
        """Test PKD sample rate is 40 Hz."""
        assert parkfield_run_ts_pkd.sample_rate == 40.0

    def test_pkd_data_length(self, parkfield_run_ts_pkd):
        """Test PKD run has expected data length."""
        # 2 hours at 40 Hz = 288000 samples
        assert parkfield_run_ts_pkd.dataset.time.shape[0] == 288000

    def test_pkd_time_range(self, parkfield_run_ts_pkd):
        """Test PKD data covers expected time range."""
        start_time = str(parkfield_run_ts_pkd.start)
        end_time = str(parkfield_run_ts_pkd.end)

        assert "2004-09-28" in start_time
        assert "2004-09-28" in end_time

    def test_kernel_dataset_ss_structure(self, parkfield_kernel_dataset_ss):
        """Test single-station kernel dataset has expected structure."""
        # KernelDataset has a df attribute that is a DataFrame
        assert "station" in parkfield_kernel_dataset_ss.df.columns
        assert "PKD" in parkfield_kernel_dataset_ss.df["station"].values

    def test_kernel_dataset_rr_structure(self, parkfield_kernel_dataset_rr):
        """Test RR kernel dataset has expected structure."""
        # KernelDataset has a df attribute that is a DataFrame
        assert "station" in parkfield_kernel_dataset_rr.df.columns
        stations = set(parkfield_kernel_dataset_rr.df["station"].values)
        assert "PKD" in stations
        assert "SAO" in stations


# ============================================================================
# Numerical Validation Tests
# ============================================================================


class TestParkfieldNumericalValidation:
    """Test numerical properties of processed results."""

    @pytest.fixture(scope="class")
    def processed_tf_validation(self, parkfield_kernel_dataset_ss):
        """Process transfer function for validation tests."""
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(
            parkfield_kernel_dataset_ss,
            estimator={"engine": "RME"},
            output_channels=["ex", "ey"],
        )
        return process_mth5(
            config,
            parkfield_kernel_dataset_ss,
            units="MT",
            show_plot=False,
        )

    def test_transfer_function_is_finite(
        self, processed_tf_validation, disable_matplotlib_logging
    ):
        """Test that computed transfer function contains no NaN or Inf."""
        tf_cls = processed_tf_validation

        # Check that transfer function values are finite for impedance elements
        # tf_cls.transfer_function is now a DataArray with (period, output, input)
        # Output includes ex, ey, and hz. Hz (tipper) may be NaN.
        if hasattr(tf_cls, "transfer_function"):
            tf_data = tf_cls.transfer_function
            # Check only ex and ey outputs (first 2), not hz (index 2)
            impedance_data = tf_data.sel(output=["ex", "ey"])
            assert np.all(np.isfinite(impedance_data.data))

    def test_transfer_function_shape(
        self, processed_tf_validation, disable_matplotlib_logging
    ):
        """Test that transfer function has expected shape."""
        tf_cls = processed_tf_validation

        # Transfer function should have shape (periods, output_channels, input_channels)
        if hasattr(tf_cls, "transfer_function"):
            tf_data = tf_cls.transfer_function
            # Should have dimensions: period, output, input
            assert tf_data.dims == ("period", "output", "input")
            # Output includes ex, ey, hz even though we only requested ex, ey
            assert tf_data.shape[1] == 3  # 3 output channels (ex, ey, hz)
            assert tf_data.shape[2] == 2  # 2 input channels (hx, hy)

    def test_processing_runs_without_errors(
        self, processed_tf_validation, disable_matplotlib_logging
    ):
        """Test that RR processing completes without raising exceptions."""
        # Reuse the same processed TF - just verify it exists
        tf_cls = processed_tf_validation

        assert tf_cls is not None



"""

This module contains the base class for regression functions. It follows Gary Egbert's EMTF Matlab code TRegression.m in which can be found in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes

This class originally used numpy arrays to make adapting the Matlab easier, but
experimental support for xarray is now added (2024).

"""
from typing import Optional, Union

import numpy as np
import xarray as xr
from loguru import logger

from aurora.transfer_function.regression.iter_control import IterControl


class RegressionEstimator(object):
    """
    Abstract base class for solving Y = X*b + epsilon for b, complex-valued

    Many of the robust transfer estimation methods we will use repeat the
    model of solving Y = X*b +epsilon for b.  X is variously called the "input",
    "predictor", "explanatory", "confounding", "independent" "exogenous", variable(s)
    or the "design matrix", "model matrix" or "regressor matrix".
    Y are variously called the "output", "predicted", "outcome",
    "response", "endogenous", "regressand", or "dependent" variable.  I will try to
    use input and output.

    When we "regress Y on X", we use the values of variable X to predict
    those of Y.

    Typically operates on single "frequency_band".
    Allows multiple columns of Y, but estimates b for each column separately.


    Estimated signal and noise covariance matrices can be used to compute error
    together to compute covariance for the matrix of regression coefficients b

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section. Alternatively, attributes may be documented
    inline with the attribute's declaration (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.



    Attributes
    ----------
    _X : Union[xr.Dataset, xr.DataArray, np.ndarray]
        The underlying dataset is assumed to be if shape nCH x nObs (normally 2-dimensional)
        These are the input variables.  In the matlab codes each observation
        corresponds to a row and each parameter (channel) is a column.
        These data are transposed before regression
    X : numpy array (normally 2-dimensional)
        This is a "pure array" representation of _X used to emulate Gary
        Egbert's matlab codes. It may or may not be deprecated.
    _Y : xarray.Dataset
        These are the output variables, arranged same as X above.
    Y : numpy array (normally 2-dimensional)
        This is a "pure array" representation of _Y used to emulate Gary
        Egbert's matlab codes. It may or may not be deprecated.
    b : numpy array (normally 2-dimensional)
        Matrix of regression coefficients, i.e. the solution to the regression
        problem.  In our context they are the "Transfer Function"
        The dimension of b, to match X and Y is [n_input_ch, n_output_ch]
        When we are solving "channel-by-channel", b is usually [2,1].

    inverse_signal_covariance: numpy array (n_input_ch, n_input_ch)
        This was Cov_SS in Gary's matlab codes.  It is basically inv(X.H @ X)
        Reference needed
    noise_covariance : numpy array (n_output_ch, n_output_ch)
        This was Cov_NN in Gary's matlab codes
        Reference needed
    squared_coherence: numpy array (n_output_ch)
        This was R2 in Gary's matlab codes
        Formula?  Reference?
        Squared coherence (top row is using raw data, bottom cleaned, with crude
        correction for amount of downweighted data)
    Yc : numpy array (normally 2-dimensional)
        A "cleaned" version of Y the output variables.
    iter_control : transfer_function.iter_control.IterControl()
        is a structure which controls the robust scheme iteration.
        On return also contains number of iterations.

    Kwargs:
    input_channel_names: list
        List of strings for channel names.

    """

    def __init__(
        self,
        X: Union[xr.Dataset, xr.DataArray, np.ndarray],
        Y: Union[xr.Dataset, xr.DataArray, np.ndarray],
        iter_control: IterControl = IterControl(),
        input_channel_names: Optional[Union[list, None]] = None,
        output_channel_names: Optional[Union[list, None]] = None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        X: Union[xr.Dataset, xr.DataArray, np.ndarray]
            The input channels for regression
        Y: Union[xr.Dataset, xr.DataArray, np.ndarray]
            The output channels for regression
        iter_control: IterControl
            Contains parameters controlling the regression, e.g. convergence criteria, thresholds, etc.
        input_channel_names: Optional[Union[list, None]]
            If X is np.ndarray, this allows associating channel names to X's columns
        output_channel_names: Optional[Union[list, None]]
            If Y is np.ndarray, this allows associating channel names to Y's columns

        """
        self._X = X
        self._Y = Y
        self.iter_control = iter_control
        self.b = None
        self.cov_nn = None
        self.cov_ss_inv = None
        self.squared_coherence = None
        self._set_up_regression_variables()
        self.R2 = None
        self.qr_input = "X"
        self._Q = None
        self._R = None
        self._QH = None  # conjugate transpose of Q (Hermitian operator)
        self._QHY = None  #
        self._QHYc = None
        self._n_channels_out = None
        self._input_channel_names = input_channel_names
        self._output_channel_names = output_channel_names

    def _set_up_regression_variables(self) -> None:
        """
        Initialize arrays needed for regression and cast any xarray to numpy

        **Development Notes**:

        When self._X,_Y are xr.Datasets they are cast to arrays (num channels x num observations) and then transposed.
        When self._X,_Y are xr.DataArrays we extract the array -- but how do we know whether or not to transpose?  It would be nice to have a helper function that applies the logic of getting the data from the xarray and transposing or not appropriately. For now we assume that the input data are organized so that input arrays are (n_ch x n_observations). This assumption is OK for xr.Dataset where the datavars are the MT components ("hx", "hy", etc)

        """
        self.X = _input_to_numpy_array(self._X)
        self.Y = _input_to_numpy_array(self._Y)
        self.Yc = np.zeros(self.Y.shape, dtype=np.complex128)
        self._check_number_of_observations_xy_consistent()

    @property
    def input_channel_names(self) -> list:
        """returns the list of channel names associated with X"""
        if self._input_channel_names is None:
            self._input_channel_names = _get_channel_names(self._X, label="IN")
        return self._input_channel_names

    @property
    def output_channel_names(self) -> list:
        """returns the list of channel names associated with Y"""
        if self._output_channel_names is None:
            self._output_channel_names = _get_channel_names(self._Y, label="OUT")
        return self._output_channel_names

    @property
    def inverse_signal_covariance(self) -> np.ndarray:
        """Returns the inverse signal covariance matrix of the input channels as xarray"""
        return self.cov_ss_inv

    @property
    def noise_covariance(self) -> xr.DataArray:
        """Returns the noise covariance matrix of the output channels  as xarray"""
        return self.cov_nn

    def b_to_xarray(self) -> xr.DataArray:
        """Wraps the TF results as an xarray"""
        xra = xr.DataArray(
            np.transpose(self.b),
            dims=["output_channel", "input_channel"],
            coords={
                "output_channel": self.output_channel_names,
                "input_channel": self.input_channel_names,
            },
        )
        return xra

    def solve_underdetermined(self) -> None:
        """
        Solves the regression problem if it is under-determined -- Not Stable

        Development Notes:
        20210806
        This method was originally in TRME.m, but it does not depend in
        general on using RME method so putting it in the base class.

        We basically never get here and when we do, we don't trust the results
        https://docs.scipy.org/doc/numpy-1.9.2/reference/generated/numpy.linalg.svd.html
        https://www.mathworks.com/help/matlab/ref/double.svd.html

        Note that the svd outputs are different in matlab and numpy
        https://stackoverflow.com/questions/50930899/svd-command-in-python-v-s-matlab
        "The SVD of a matrix can be written as

        A = U S V^H

        Where the ^H signifies the conjugate transpose. Matlab's svd command returns U, S and V,
        while numpy.linalg.svd returns U, the diagonal of S, and V^H.
        Thus, to get the same S and V as in Matlab you need to reconstruct the S and also get the V:


        ORIGINAL MATLAB

        % Overdetermined problem...use svd to invert, return
        % NOTE: the solution IS non - unique... and by itself RME is not setup
        % to do anything sensible to resolve the non - uniqueness(no prior info
        % is passed!).  This is stop-gap, to prevent errors when using RME as
        % part of some other estimation scheme!
        [u,s,v] = svd(obj.X,'econ');
        sInv = 1./diag(s);
        obj.b = v*diag(sInv)*u'*obj.Y;
        if ITER.returnCovariance
           obj.Cov_NN = zeros(K,K);
           obj.Cov_SS = zeros(nParam,nParam);
        end
        result = obj.b


        -------

        """
        U, s, V = np.linalg.svd(self.X, full_matrices=False)
        S_inv = np.diag(1.0 / s)
        self.b = (V.T.conj() @ S_inv @ U.T.conj()) * self.Y
        if self.iter_control.return_covariance:
            logger.warning(
                "Warning covariances are not xarray, may break things downstream"
            )
            self.cov_nn = np.zeros((self.n_channels_out, self.n_channels_out))
            self.cov_ss_inv = np.zeros((self.n_channels_in, self.n_channels_in))

        return

    def _check_number_of_observations_xy_consistent(self) -> None:
        """Raises an exception if the X, Y data have different number of observations"""
        if self.Y.shape[0] != self.X.shape[0]:
            logger.error(
                f"Design matrix (X) has {self.X.shape[0]} rows but data (Y) "
                f"has {self.Y.shape[0]}"
            )
            raise Exception

    @property
    def n_data(self) -> int:
        """
        Return the number of multivariate observations in the regression dataset.

        Development Notes:
         This may need to be modified if we use masked arrays

        Returns
        -------
        int:
            The number of rows in the input data vector
        """
        return self.X.shape[0]

    @property
    def n_channels_in(self) -> int:
        """
        Returns the number of input channels.

        Returns
        -------
        int:
            The number of channels of input data (columns of X)
        """
        return self.X.shape[1]

    @property
    def n_channels_out(self) -> int:
        """
        number of output variables

        Returns
        int
            number of output variables (Assumed to be num columns of a 2D array)
        """
        if self._n_channels_out is None:
            self._n_channels_out = self.Y.shape[1]
        return self._n_channels_out

    @property
    def degrees_of_freedom(self) -> int:
        """
        Gets the number of degrees of freedom in the dataset.

        Returns
        -------
        int
            The total number of multivariate observations minus the number of input channels.
        """
        return self.n_data - self.n_channels_in

    @property
    def is_underdetermined(self) -> bool:
        """
        Check if the regression problem is under-determined
        Returns
        -------
        bool
            True if regression problem is under-determined, otherwise False
        """
        return self.degrees_of_freedom < 0

    # TODO Add support for masked arrays
    # def mask_input_channels(self):
    #     """
    #     ADD NAN MANAGEMENT HERE
    #     Returns
    #     -------
    #
    #     """
    #     pass

    def qr_decomposition(
        self, X: Optional[Union[np.ndarray, None]] = None, sanity_check: bool = False
    ) -> tuple:
        """
        performs QR decomposition on input matrix X.

        If X is not provided as an optional argument then check the value of self.qr_input

        Parameters
        ----------
        X: numpy array
            In RME this is the Input channels X
            In RME_RR this is the RR channels Z
        sanity_check: boolean
            check QR decomposition is working correctly.  Set to True for debugging.
            Can probably be deprecated.

        Returns
        -------
        Q, R: tuple
            The matrices Q and R from the QR-decomposition.
            X = Q R where Q is unitary/orthogonal and R upper triangular.
        """
        if X is None:
            if self.qr_input == "X":
                X = self.X
            elif self.qr_input == "Z":
                X = self.Z
            else:
                msg = f"Matrix to perform QR decomposition not specified by {self.qr_input}"
                logger.error("msg")
                raise ValueError(msg)

        Q, R = np.linalg.qr(X)
        self._Q = Q
        self._R = R
        if sanity_check:
            if np.isclose(np.matmul(Q, R) - X, 0).all():
                pass
            else:
                logger.error("Failed QR decomposition sanity check")
                raise Exception
        return Q, R

    @property
    def Q(self) -> np.ndarray:
        """Returns the Q matrix from the QR decomposition"""
        return self._Q

    @property
    def R(self) -> np.ndarray:
        """Returns the R matrix from the QR decomposition"""
        return self._R

    @property
    def QH(self) -> np.ndarray:
        """Returns the conjugate transpose of the Q matrix from the QR decomposition"""
        if self._QH is None:
            self._QH = self.Q.conj().T
        return self._QH

    @property
    def QHY(self) -> np.ndarray:
        """
        Returns the QH @ Y

        QHY is a convenience matrix that makes computing the predicted data (QQHY) more efficient.

        """
        if self._QHY is None:
            self._QHY = self.QH @ self.Y
        return self._QHY

    def estimate_ols(self, mode: Optional[str] = "solve") -> np.ndarray:
        """
        Solve Y = Xb with ordinary least squares, not robust regression.

        Development Notes:
        Brute Force tends to be less stable because we actually compute the
        inverse matrix.  It is not recommended, its just here for completeness.
        X'Y=X'Xb
        (X'X)^-1 X'Y = b

        Parameters
        ----------
        mode : str
            must be one of ["qr", "brute_force", "solve"]

        Returns
        -------
        b : numpy array
            Normally the impedance tensor Z

        """
        if mode.lower() == "qr":
            from scipy.linalg import solve_triangular

            self.qr_decomposition(self.X)
            b = solve_triangular(self.R, self.QH @ self.Y)
        else:
            X = self.X
            Y = self.Y
            XH = np.conj(X.T)
            XHX = XH @ X
            XHY = XH @ Y
            if mode.lower() == "brute_force":
                XHX_inv = np.linalg.inv(XHX)
                b = np.matmul(XHX_inv, XHY)
            elif mode.lower() == "solve":
                b = np.linalg.solve(XHX, XHY)
            else:
                logger.error(f"mode {mode} not recognized")
                raise Exception
        self.b = b
        return b

    def estimate(self) -> np.ndarray:
        """
        Executes the regression

        Returns
        -------
        b: np.ndarray
            The regression solution to Y = Xb
        """
        b = self.estimate_ols(mode="qr")
        return b


def _input_to_numpy_array(X: Union[xr.Dataset, xr.DataArray, np.ndarray]) -> np.ndarray:
    """
    Casts data input to regression as numpy array, with channels as column vectors.

    Currently, we store array channels row-wise (as num_channels x num_observations),
    but the way regression is set up the variables should be column vectors (num_observations x num_channels)

    This is a place where we could distill the logic for which dimension is which.


    Parameters
    ----------
    X: Union[xr.Dataset, xr.DataArray]
        Data to be used in regression in xarray form

    Returns
    -------
    output: np.ndarray
        Data to be used in regression as a numpy array

    """
    if isinstance(X, xr.Dataset):
        output = X.to_array().data.T
    elif isinstance(X, xr.DataArray):
        output = X.data.T
        if len(output.shape) == 1:
            output = np.atleast_2d(output).T  # cast to 2D if 1D
    elif isinstance(X, np.ndarray):
        msg = "np.ndarray input is assumed to be nCH x nObs -- transposing"
        logger.debug(msg)
        output = X.T
    else:
        msg = f"input argument of type {type(X)} not supported -- try an xarray"
        raise NotImplementedError(msg)

    return output


def _get_channel_names(
    X: Union[xr.Dataset, xr.DataArray, np.ndarray], label: Optional[str] = ""
) -> list:
    """
    Returns list of channel names.

    If X is a numpy array, names will be created. These are needed by TRME.estimate() to return xarrays.

    Parameters
    ----------
    X: Union[xr.Dataset, xr.DataArray, np.ndarray]
        If X is xarray just return the labels
        If X is numpy array, make the names up.  numpy array assumed to contain data from each channel
        in a separate row, i.e. (n_ch x n_observations) shaped array.
    label: Optional[str]
        This gets prepended onto incrementing integers for channel labels.
        For example, this could be "input", "output", or a station name.
        Used to keep the indexing a 2D xarray unique.

    Returns
    -------
    channel_names: list
        The names of the channels for the input array X
    """
    if isinstance(X, xr.Dataset):
        channel_names = list(X.data_vars)
    elif isinstance(X, xr.DataArray):
        # Beware hard coded assumption of "variable"
        try:
            channel_names = list(X.coords["variable"].values)
        except TypeError:  # This happens when xarray has only one channel
            channel_names = [
                X.coords["variable"].values.item(),
            ]
    else:
        # numpy array doesn't have input channel_names predefined
        channel_names = np.arange(
            X.shape[0]
        )  # note its 0, not 1 here because we are talking to _X
        channel_names = [f"{label}{x}" for x in channel_names]
    return channel_names



"""
This is a sandbox folder.  Probably most of what is here will wind up in mt_metadata, mtpy, or mth5.

While dealing with practical implementation of coherence sorting, it becomes clear
 that managing channel and feature nomenclature can become tricky.

We need a way to express the channel pairs that cohernce will be evaluated on.

The normal channel pairs are:
    ex-hy
    ey-hx
    ex-ref_ex
    ey-ref_ey
    hx-ref_hx
    hy-ref_hy

    where "ref_" refers to the remote reference station.

For single station processing, only the first two are relevant.
For remote reference, all six are relevant, with usage of ref_ex, ref_ey
 being __far__ less common that the magnetic channels at the remote station.

It seems useful to have the notion of a "remote reference context",
 where we can tell a processing configuration to use for example,

ey-hx
hx-ref_hx

as our coherece sorting channels, without spelling out the reference
 channel name explictly.  This is much like what happens inside
 transfer_function_helpers.process_transfer_functions, which uses the nomenclature X, Y, RR.

Within this "remote reference context", we can compute/access the appropriate coherences.

This can be revisited later in a multivariate context.

"""



r"""
This module is a place to put decorators used by methods in aurora.
It is a work in progress.

Development notes:

Here is the decorator pattern::

    def decorator(func):
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            # Do something before
            value = func(*args, **kwargs)
            # Do something after
            return value
        return wrapper_decorator

"""

import functools

import xarray as xr


def can_use_xr_dataarray(func):
    """
    Decorator to allow functions that operate on xr.Dataset to also operate
    on xr.DataArray, by converting xr.DataArray to xr.Dataset, calling the
    function, and then  converting back to xr.DataArray

    Most of the windowed time series methods are
    intended to work with xarray.Dataset class.  But I would like to be able
    to pass them xarray.DataArray objects.  This class casts a DataArray to a
    Dataset, runs it through func and casts back to a DataArray.

    A similar decorator could be written for numpy arrays.
    Parameters
    ----------
    func

    Returns
    -------

    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        if isinstance(kwargs["data"], xr.DataArray):
            kwargs["data"] = kwargs["data"].to_dataset("channel")
            input_was_dataarray = True
        else:
            input_was_dataarray = False

        if isinstance(func, staticmethod):
            callable_func = func.__get__(None, object)
            processed_obj = callable_func(*args, **kwargs)
        else:
            processed_obj = func(*args, **kwargs)

        if input_was_dataarray:
            processed_obj = processed_obj.to_array("channel")
        return processed_obj

    return wrapper_decorator



"""
    This module contains helper functions for working with obspy objects.

    Development Notes
    - TODO: Most of this could likely be moved into MTH5

"""
import datetime

from obspy import UTCDateTime
from loguru import logger


def trim_streams_to_common_timestamps(streams):
    """
    Slices streams to common time stamps.

    Parameters
    ----------
    streams

    Returns
    -------

    """
    start_times = sorted(list(set([tr.stats.starttime.isoformat() for tr in streams])))
    end_times = sorted(list(set([tr.stats.endtime.isoformat() for tr in streams])))
    run_stream = streams.slice(UTCDateTime(start_times[0]), UTCDateTime(end_times[-1]))
    return run_stream


def align_streams(streams, clock_start):
    """
    Shift stream timestamps so that they are aligned.

    This is a hack around to handle data that are asynchronously sampled.
    It should not be used in general.  It is only appropriate for datasets that have
    been tested with it.
    PKD, SAO only at this point.

    Parameters
    ----------
    streams : iterable of types obspy.core.stream.Stream
    clock_start : obspy UTCDateTime
        this is a reference time that we set the first sample to be

    Returns
    -------
    streams
    """
    for stream in streams:
        logger.info(
            f"{stream.stats['station']}  {stream.stats['channel']} N="
            f"{len(stream.data)}  startime {stream.stats.starttime}"
        )
        dt_seconds = stream.stats.starttime - clock_start
        logger.info(f"stream offset seconds {dt_seconds}")
        dt = datetime.timedelta(seconds=dt_seconds)
        logger.info(f"stream timedelta {dt}")
        stream.stats.starttime = stream.stats.starttime - dt
    return streams


FDSN_CHANNEL_MAP = {}
FDSN_CHANNEL_MAP["BQ2"] = "BQ1"
FDSN_CHANNEL_MAP["BQ3"] = "BQ2"
FDSN_CHANNEL_MAP["BT1"] = "BF1"
FDSN_CHANNEL_MAP["BT2"] = "BF2"
FDSN_CHANNEL_MAP["BT3"] = "BF3"
FDSN_CHANNEL_MAP["LQ2"] = "LQ1"
FDSN_CHANNEL_MAP["LQ3"] = "LQ2"
FDSN_CHANNEL_MAP["LT1"] = "LF1"
FDSN_CHANNEL_MAP["LT2"] = "LF2"
FDSN_CHANNEL_MAP["LT3"] = "LF3"
FDSN_CHANNEL_MAP["LFE"] = "LF1"
FDSN_CHANNEL_MAP["LFN"] = "LF2"
FDSN_CHANNEL_MAP["LFZ"] = "LF3"
FDSN_CHANNEL_MAP["LQE"] = "LQ1"
FDSN_CHANNEL_MAP["LQN"] = "LQ2"


def make_channel_labels_fdsn_compliant(streams):
    """
    Renames channels to FDSN compliant channel names.

    Workaround because NCEDC channel nomenclature is not FDSN Compliant for PKD, SAO
    Specifically, re-assign non-conventional channel labels
    Q2 --> Q1
    Q3 --> Q2
    T1 --> F1
    T2 --> F2
    T3 --> F3

    Parameters
    ----------
    streams : iterable of types obspy.core.stream.Stream

    Returns
    -------
    streams : iterable of types obspy.core.stream.Stream
        Same as input but updated with new channel names.
    """
    for stream in streams:
        stream.stats["channel"] = FDSN_CHANNEL_MAP[stream.stats["channel"]]
    return streams



"""
One off method to help read in transfer function dumps provided by Gary from some of
the matlab tests.

"""

import numpy as np
import pandas as pd
import scipy.io as sio

from aurora.config.metadata.processing import Processing
from aurora.config.emtf_band_setup import BANDS_256_29_FILE
from aurora.general_helper_functions import get_test_path
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.sandbox.io_helpers.emtf_band_setup import EMTFBandSetupFile
from aurora.transfer_function.emtf_z_file_helpers import clip_bands_from_z_file

from aurora.transfer_function.transfer_function_collection import (
    TransferFunctionCollection,
)
from mt_metadata.timeseries.survey import Survey
from mt_metadata.transfer_functions.core import TF
from loguru import logger

from mth5.processing import KernelDataset

TEST_PATH = get_test_path()


def read_matlab_z_file(case_id, z_mat):
    tmp = sio.loadmat(z_mat)
    if case_id == "synthetic":
        tmp = tmp["temp"][0][0].tolist()
    elif case_id == "IAK34ss":
        tmp = tmp["TFstruct"][0][0].tolist()
    return tmp


def test_matlab_zfile_reader(case_id="IAK34ss", make_plot=False):
    """

    Parameters
    ----------
    case_id: string
        one of ["IAK34ss", "synthetic"]
        currently only IAK34ss is supported
    make_plot: bool
        Set to True when debugging

    Takes a stored matlab file: "IAK34_struct_zss.mat", reads it in, and packs
    a z-file from it.

    There is a stored version of the z-file from July 2022 (archived in git in Sept
    2022) that we assert equality with to pass the test.

    ToDo: There is another stored mat file:"TS1zss20210831.mat"
    this appears to be synthetic station test1, processed with BANDS_256_29_FILE
    The output z-file from this can be compared to test1_aurora_matlab.zss.
    The results should agree approximately.


    """
    test_dir_path = TEST_PATH.joinpath("io")

    # 1. Define processing parameters based on which of the two tests are to be run
    if case_id == "synthetic":
        band_setup_file = BANDS_256_29_FILE
        field_data_sample_rate = 1.0
        num_samples_window = 256
        decimation_factors = [1, 4, 4, 4]
        n_periods_clip = 3  # for synthetic case
        estimator_engine = "RME"
        local_station_id = "test1"
        remote_station_id = ""
        input_channels = ["hx", "hy"]
        output_channels = [
            "hz",
            "ex",
            "ey",
        ]
        reference_channels = []
        matlab_z_file = "TS1zss20210831.mat"
        archived_z_file_path = None
        z_file_path = "from_matlab.zss"
    elif case_id == "IAK34ss":
        band_setup_file = BANDS_256_29_FILE
        field_data_sample_rate = 1.0
        num_samples_window = 256
        decimation_factors = [1, 4, 4, 4]
        n_periods_clip = 3
        estimator_engine = "RME"
        local_station_id = "IAK34"
        remote_station_id = ""
        input_channels = ["hx", "hy"]
        output_channels = [
            "hz",
            "ex",
            "ey",
        ]
        reference_channels = []
        matlab_z_file = test_dir_path.joinpath("IAK34_struct_zss.mat")
        archived_z_file_path = test_dir_path.joinpath("archived_from_matlab.zss")
        z_file_path = test_dir_path.joinpath("from_matlab.zss")

    # 2. Create an aurora processing config
    p = Processing()
    p.stations.local.id = local_station_id
    if remote_station_id:
        p.stations.remote.id = remote_station_id
    emtf_band_setup = EMTFBandSetupFile(
        filepath=band_setup_file, sample_rate=field_data_sample_rate
    )
    band_edges = emtf_band_setup.compute_band_edges(
        decimation_factors=decimation_factors,
        num_samples_window=4 * [num_samples_window],
    )
    p.assign_bands(
        band_edges,
        field_data_sample_rate,
        decimation_factors,
        num_samples_window,
    )

    # 3. populate decimation levels of processing config
    tf_dict = {}
    sample_rate = field_data_sample_rate
    for i_dec in range(4):
        p.decimations[i_dec].decimation.sample_rate = sample_rate
        p.decimations[i_dec].stft.window.num_samples = num_samples_window
        p.decimations[i_dec].estimator.engine = estimator_engine
        p.decimations[i_dec].input_channels = input_channels
        p.decimations[i_dec].output_channels = output_channels
        p.decimations[i_dec].reference_channels = reference_channels

        tffz_obj = p.make_tf_level(i_dec)
        tf_dict[i_dec] = tffz_obj
        sample_rate /= 4.0

    # 4. Read in the matlab_z object
    tmp = read_matlab_z_file(case_id, matlab_z_file)
    matlab_tf = tmp[4]
    periods = tmp[5]
    cov_ss = tmp[7]
    cov_nn = tmp[8]
    n_data = tmp[9]
    R2 = tmp[10]
    # stderr = tmp[12]
    # freq = tmp[14]

    # 4a. FIX NAN / INF
    nan_cov_nn = []
    for i in range(len(periods)):
        if np.isnan(cov_nn[:, :, i]).any():
            nan_cov_nn.append(i)
            logger.info(f"NAN {i}")

    if case_id == "synthetic":
        cov_nn[:, :, 28] = cov_nn[:, :, 27]
    elif case_id == "IAK34ss":
        for i in range(12):
            cov_nn[:, :, i] = cov_nn[:, :, 12]

    # 5. Pack the TF dicts with data values
    tf_dict[0].tf.data = matlab_tf[:, :, :11]
    tf_dict[0].cov_nn.data = cov_nn[:, :, :11]
    tf_dict[0].cov_ss_inv.data = cov_ss[:, :, :11]
    tf_dict[0].num_segments.data = n_data[:, :11]
    tf_dict[0].R2.data = R2[:, :11]

    tf_dict[1].tf.data = matlab_tf[:, :, 11:17]
    tf_dict[1].cov_nn.data = cov_nn[:, :, 11:17]
    tf_dict[1].cov_ss_inv.data = cov_ss[:, :, 11:17]
    tf_dict[1].num_segments.data = n_data[:, 11:17]
    tf_dict[1].R2.data = R2[:, 11:17]

    tf_dict[2].tf.data = matlab_tf[:, :, 17:23]
    tf_dict[2].cov_nn.data = cov_nn[:, :, 17:23]
    tf_dict[2].cov_ss_inv.data = cov_ss[:, :, 17:23]
    tf_dict[2].num_segments.data = n_data[:, 17:23]
    tf_dict[2].R2.data = R2[:, 17:23]

    tf_dict[3].tf.data = matlab_tf[:, :, 23:]
    tf_dict[3].cov_nn.data = cov_nn[:, :, 23:]
    tf_dict[3].cov_ss_inv.data = cov_ss[:, :, 23:]
    tf_dict[3].num_segments.data = n_data[:, 23:]
    tf_dict[3].R2.data = R2[:, 23:]

    # 6. Create TF Collection, and TFKernel
    tfc = TransferFunctionCollection(tf_dict=tf_dict, processing_config=p)
    kd = KernelDataset(local_station_id=local_station_id)
    survey_metadata = Survey()
    kd.survey_metadata["0"] = survey_metadata
    kd_df_dict = {
        "remote": [False],
        "station": [local_station_id],
        "processing_type": ["matlab EMTF"],
        "survey": ["0"],
        "run": ["a"],
        "start": ["1980-01-01T00:00:00"],
        "end": ["1980-01-02T00:00:00"],
    }
    kd_df = pd.DataFrame(data=kd_df_dict)
    kd.df = kd_df
    tfk = TransferFunctionKernel(dataset=kd, config=p)

    # 7. Create mt_metadata TF object and write z-file
    tf_obj = tfk.export_tf_collection(tfc)
    tf_obj.station_metadata.id = local_station_id
    tf_obj.station_metadata.runs[0].channels["ey"].measurement_azimuth = 90.0
    tf_obj.station_metadata.runs[0].channels["hy"].measurement_azimuth = 90.0
    tf_obj.write(z_file_path)

    # 7a. clip last few periods (matlab object and z-objects provided were not identical)
    if n_periods_clip:
        clip_bands_from_z_file(z_file_path, n_periods_clip, n_sensors=5)

    # 8. Compare z-file written by TF object to the archived version
    tf_obj_from_z_file = TF()
    tf_obj_from_z_file.from_zmm(z_file_path)
    archived_tf = TF()
    archived_tf.from_zmm(archived_z_file_path)
    assert np.isclose(
        tf_obj_from_z_file.transfer_function.data,
        archived_tf.transfer_function.data,
        rtol=1e-3,
    ).all()

    logger.info("success!")



"""
    This module contains functions to used by test that involve Fourier Coefficients in MTH5.
"""

from typing import Optional, Union
import numpy as np
import pandas as pd
import pathlib
import xarray as xr


def read_fc_csv(
    csv_name: Union[pathlib.Path, str], as_xarray: Optional[bool] = True
) -> Union[xr.Dataset, pd.DataFrame]:
    """

    Load Fourier coefficients from a csv file and return as xarray or dataframe

    Usage:
    xrds_obj = read_fc_csv(csv_name)
    df = read_fc_csv(csv_name, as_xarry=False)

    Parameters
    ----------
    csv_name: Union[pathlib.Path, str]
        Path to csv file to read
    as_xarray: Optional[bool]
        If true return xr.Dataset

    Returns
    -------
    output: xr.Dataset or pd.DataFrame
    """
    df = pd.read_csv(
        csv_name,
        index_col=[0, 1],
        parse_dates=[
            "time",
        ],
        skipinitialspace=True,
    )
    for col in df.columns:
        df[col] = np.complex128(df[col])
    if as_xarray:
        output = df.to_xarray()
    else:
        output = df
    return output



"""
This module contains an extension of aurora's TransferFunction base class.
This class can return estimates of standard error, apparent resistivity and phase.

Development Notes:
This class follows  Gary's legacy matlab code  TTFZ.m from
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
TODO: This should be replaced by methods in mtpy.
"""

import numpy as np
import xarray as xr
from loguru import logger
from matplotlib import pyplot as plt

from aurora.transfer_function.base import TransferFunction


class TTFZ(TransferFunction):
    """
    subclass to support some more MT impedance specific functions  --
    initially just apparent resistivity and phase for diagonal elements.
    + rotation/fixed coordinate system

    TODO: This class should be deprecated and mt_metadata TF object should be used instead.

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: passed through to base class
        kwargs: passed through to base class
        """
        super(TTFZ, self).__init__(*args, **kwargs)

    def standard_error(self):
        """
        estimate the standard error, used for error bars and inversion.

        Development Notes:
        The standard error is normally thought of as the sqrt of the error variance.
        since the code here sets std_err = np.sqrt(np.abs(cov_ss_inv * cov_nn))
        that means the inverse signal covariance times the noise covariance is like the error variance.

        Returns
        -------
        standard_error: xr.DataArray
        """
        stderr = np.zeros(self.tf.data.shape)
        standard_error = xr.DataArray(
            stderr,
            dims=["output_channel", "input_channel", "period"],
            coords={
                "output_channel": self.tf_header.output_channels,
                "input_channel": self.tf_header.input_channels,
                "period": self.periods,
            },
        )
        for out_ch in self.tf_header.output_channels:
            for inp_ch in self.tf_header.input_channels:
                for T in self.periods:
                    cov_ss_inv = self.cov_ss_inv.loc[inp_ch, inp_ch, T]
                    cov_nn = self.cov_nn.loc[out_ch, out_ch, T]
                    std_err = np.sqrt(np.abs(cov_ss_inv * cov_nn))
                    standard_error.loc[out_ch, inp_ch, T] = std_err

        return standard_error

    def apparent_resistivity(self, channel_nomenclature, units="SI"):
        """
        Computes the apparent resistivity and phase.

        Development notes:
        Original Matlab Documentation:
        ap_res(...) : computes app. res., phase, errors, given imped., cov.
        %USAGE: [rho,rho_se,ph,ph_se] = ap_res(z,sig_s,sig_e,periods) ;
        % Z = array of impedances (from Z_***** file)
        % sig_s = inverse signal covariance matrix (from Z_****** file)
        % sig_e = residual covariance matrix (from Z_****** file)
        % periods = array of periods (sec)

        Parameters
        ----------
        units: str
            one of ["MT","SI"]
        channel_nomenclature:
        mt_metadata.processing.aurora.channel_nomenclature.ChannelNomenclature
            has a dict that maps the channel names in TF to the standard channel labellings.

        """
        ex, ey, hx, hy, hz = channel_nomenclature.unpack()
        rad_deg = 180 / np.pi
        # off - diagonal impedances
        self.rho = np.zeros((self.num_bands, 2))
        self.rho_se = np.zeros((self.num_bands, 2))
        self.phi = np.zeros((self.num_bands, 2))
        self.phi_se = np.zeros((self.num_bands, 2))
        Zxy = self.tf.loc[ex, hy, :].data
        Zyx = self.tf.loc[ey, hx, :].data

        # standard deviation of real and imaginary parts of impedance
        Zxy_se = self.standard_error().loc[ex, hy, :].data / np.sqrt(2)
        Zyx_se = self.standard_error().loc[ey, hx, :].data / np.sqrt(2)

        if units == "SI":
            rxy = 2e-7 * self.periods * (abs(Zxy) ** 2)
            ryx = 2e-7 * self.periods * (abs(Zyx) ** 2)
            # print("Correct the standard errors for SI units")
            Zxy_se *= 1e-3
            Zyx_se *= 1e-3
            rxy_se = 2 * np.sqrt(self.periods * rxy / 5) * Zxy_se
            ryx_se = 2 * np.sqrt(self.periods * ryx / 5) * Zyx_se
        elif units == "MT":
            rxy = 2e-1 * self.periods * (abs(Zxy) ** 2)
            ryx = 2e-1 * self.periods * (abs(Zyx) ** 2)
            rxy_se = 2 * np.sqrt(self.periods * rxy / 5) * Zxy_se
            ryx_se = 2 * np.sqrt(self.periods * ryx / 5) * Zyx_se
        else:
            logger.error("ERROR: only SI and MT units supported")
            raise Exception

        self.rho[:, :] = np.vstack((rxy, ryx)).T
        self.rho_se[:, :] = np.vstack((rxy_se, ryx_se)).T

        # phases
        pxy = rad_deg * np.arctan(np.imag(Zxy) / np.real(Zxy))
        pyx = rad_deg * np.arctan(np.imag(Zyx) / np.real(Zyx))
        self.phi[:, :] = np.vstack((pxy, pyx)).T

        pxy_se = rad_deg * Zxy_se / np.abs(Zxy)
        pyx_se = rad_deg * Zyx_se / np.abs(Zyx)

        self.phi_se = np.vstack((pxy_se, pyx_se)).T
        return

    def plot(self, out_filename=None, **kwargs):
        """Plot the transfer function using mtpy's built in plot function."""

        plot_object = RhoPlot(self)
        plt.ion()
        return plot_object.plot(
            station_id=self.tf_header.local_station.id,
            out_filename=out_filename,
            **kwargs,
        )


plt.ioff()


class RhoPlot(object):
    """
    TF plotting object class; some methods are only relevant to
    specific types of TFs (or for derived parameters such as rho/phi)

    Development Notes:
        This should be deprecated and replaced with MTpy
        The only place this class is used is in aurora/sandbox/plot_helpers.py in the
        plot_tf_obj method.

    """

    def __init__(self, tf_obj):
        """
        Constructor

        TODO: Replace tf_obj with mt_metadata tf if this method not replaced with mtpy.

        Parameters
        ----------
        tf_obj: aurora.transfer_function.TTFZ.TTFZ
            Object with TF information


        """
        self.tf = tf_obj
        self._blue = "steelblue"
        self._red = "firebrick"

    def err_log(
        self,
        x: np.ndarray,
        y: np.ndarray,
        yerr: np.ndarray,
        x_axis_limits: list,
        log_x_axis: bool = True,
        barsize: float = 0.0075,
    ):
        """
            Returns the coordinates for the line segments that make up the error bars.

        Development Notes:
        This function returns 6 numbers per data point.
        There is no documentation for what it does.
        A reasonable guess would be that the six numbers define 3 line segments.
        One line segment for the error bar, and one line segment at the top of the error bar, and one at the bottom.
        The vectors xb and yb each have six elements per data point assigned as follows
        xb = [x-dx, x+dx, x, x, x-dx, x+dx,]
        yb = [y-dy, y-dy, y-dy, y+dy, y+dy, y+dy,]
        and if log_x_axis is True
        [log(x)-dx, log(x)+dx, log(x), log(x), log(x)-dx, log(x)+dx,]

        Matlab Documentation
        err_log : used for plotting error bars with a y-axis log scale
        takes VECTORS x and y and outputs matrices (one row per data point) for
        plotting error bars ll = 'XLOG' for log X axis

        Parameters
        ----------
        x : np.ndarray
            The x-axis values.  Usually these are periods with units of seconds
        y : np.ndarray
            The x-axis values.  Usually apparent resistivity or phase
        yerr: np.ndarray
            A value associated with the error in the y measurement.
            It seems that this is the "half height" of the error bar.
        log_x_axis : bool
            If True the xaxis is logarithmic
            Not tested for False
        x_axis_limits: list
            The lower and upper limits for the xaxis in position 0, 1 respectively.
        barsize: float
            The width of the top and bottom horizontal error bar lines.

        Returns
        -------
        xb, yb: tuple
            Each is np.ndarray, 6 rows and one column per data point
            These are the six points needed to draw the error bars.
        """
        num_observations = len(x)
        xb = np.zeros((6, num_observations))
        yb = np.zeros((6, num_observations))
        if log_x_axis:
            dx = (
                np.log(x_axis_limits[1] / x_axis_limits[0]) * barsize
            )  # natural log in matlab & python
            xb[2, :] = np.log(x)
        else:
            dx = (x_axis_limits[1] - x_axis_limits[0]) * barsize
            xb[2, :] = x
        xb[3, :] = xb[2, :]
        xb[0, :] = xb[2, :] - dx
        xb[1, :] = xb[2, :] + dx
        xb[4, :] = xb[2, :] - dx
        xb[5, :] = xb[2, :] + dx

        if log_x_axis:
            xb = np.exp(xb)

        yb[0, :] = (y - yerr).T
        yb[1, :] = (y - yerr).T
        yb[2, :] = (y - yerr).T
        yb[3, :] = (y + yerr).T
        yb[4, :] = (y + yerr).T
        yb[5, :] = (y + yerr).T

        return xb, yb

    def phase_sub_plot(self, ax, ttl_str="", pred=None, linewidth=2):
        """
        place a phase subplot on given figure axis

        Development notes:
         Originally this took an optional input argument `axRect`
         but it was never used. It looks as it it was intended to be able to set the
         position of the figure.  There was also some hardcoded control of linewidth
         and markersize which has been removed for readability.


        Parameters
        ----------
        ax
        pred

        Returns
        -------

        """

        phi = self.tf.phi
        # rotate phases so all are positive:
        negative_phi_indices = np.where(phi < 0)[0]
        phi[negative_phi_indices] += 180.0

        Tmin, Tmax = self.set_period_limits()
        axis_limits = [Tmin, Tmax, 0, 90]

        [xb, yb] = self.err_log(
            np.transpose(self.tf.periods),
            self.tf.phi[:, 0],
            self.tf.phi_se[:, 0],
            axis_limits,
            log_x_axis=True,
        )

        ax.semilogx(xb, yb, ls="-", color=self._blue)
        ax.semilogx(self.tf.periods, phi[:, 0], marker="o", ls="--", color=self._blue)

        xb, yb = self.err_log(
            np.transpose(self.tf.periods),
            self.tf.phi[:, 1],
            self.tf.phi_se[:, 1],
            axis_limits,
            log_x_axis=True,
        )
        ax.semilogx(xb, yb, ls="-", color=self._red)
        ax.semilogx(self.tf.periods, phi[:, 1], marker="o", ls="--", color=self._red)
        # set(lines, 'LineWidth', 1, 'MarkerSize', 7);
        if pred is not None:
            plt.plot(pred.tf.periods, pred.tf.phi[:, 0], "b-")
            plt.plot(pred.tf.periods, pred.tf.phi[:, 1], "r-")

        # (lims_ph);
        ax.set_xlim(axis_limits[0], axis_limits[1])
        ax.set_ylim(axis_limits[2], axis_limits[3])

        # ax.set_subtitle( ttl_str, fontsize=14, fontweight="demi")
        # set(gca, 'FontWeight', 'bold', 'FontSize', 11, 'Xtick', xticks);
        ax.set_xlabel("Period (s)")
        ax.set_ylabel("Degrees")
        return ax

    def rho_sub_plot(self, ax, ttl_str="", pred=None):
        """
        Makes an apparent resistivity plot on the input axis.

        Matlab Documentation:
        Calls plotrhom, standard plotting routine; uses some other routines in
        EMTF/matlab/Zplt; this version is for putting multiple curves on the
        same plot ... set plotting limits now that rho is known


        Parameters
        ----------
        ax: matplotlib.axes._axes.Axes
        pred

        Returns
        -------

        """
        lims = self.set_lims()  # get the axes limits
        x_axis_limits = lims[0:2]
        y_axis_limits = lims[2:4]

        # get and plot error bars:
        [xb, yb] = self.err_log(
            self.tf.periods,
            self.tf.rho[:, 0],
            self.tf.rho_se[:, 0],
            x_axis_limits,
            log_x_axis=True,
        )
        ax.loglog(xb, yb, ls="--", color=self._blue)

        # plot rho dots
        ax.loglog(
            self.tf.periods,
            self.tf.rho[:, 0],
            marker="o",
            ls="--",
            color=self._blue,
            label="$Z_{xy}$",
        )

        [xb, yb] = self.err_log(
            self.tf.periods,
            self.tf.rho[:, 1],
            self.tf.rho_se[:, 1],
            x_axis_limits,
            log_x_axis=True,
        )
        ax.loglog(xb, yb, ls="-", color=self._red)
        ax.loglog(
            self.tf.periods,
            self.tf.rho[:, 1],
            marker="o",
            ls="--",
            color=self._red,
            label="$Z_{yx}$",
        )

        if pred is not None:
            ax.plot(
                pred.tf.periods,
                pred.tf.rho[:, 0],
                "b-",
                label="$Z_{xy}$",
            )
            ax.plot(
                pred.tf.periods,
                pred.tf.rho[:, 1],
                "r-",
                label="$Z_{yx}$",
            )

        # axis(lims_rho);
        ax.set_xlim(x_axis_limits[0], x_axis_limits[1])
        ax.set_ylim(y_axis_limits[0], y_axis_limits[1])
        ax.legend()
        ax.set_ylabel(r"$\Omega$-m")
        return ax

    def set_period_limits(self):
        """
        Returns a set of limits for the x-axis of plots based on periods to display.

        Original Matlab Notes:
            "set nicer period limits for logartihmic period scale plots"

        Returns
        -------
        Tmin, Tmax: tuple
            The minimum and maximum periods for the x-axis
        """

        x_min = self.tf.minimum_period
        x_max = self.tf.maximum_period

        Tmin = 10 ** (np.floor(np.log10(x_min) * 2) / 2)
        if (np.log10(x_min) - np.log10(Tmin)) < 0.15:
            Tmin = 10 ** (np.log10(Tmin) - 0.3)

        Tmax = 10 ** (np.ceil(np.log10(x_max) * 2) / 2)
        if (np.log10(Tmax) - np.log10(x_max)) < 0.15:
            Tmax = 10 ** (np.log10(Tmax) + 0.3)
        return Tmin, Tmax

    def set_rho_limits(self):
        """
        Returns a set of limits for the x-axis of plots based on periods to display.

        Original Matlab Notes:
            "set nicer period limits for logartihmic period scale plots"

        Returns
        -------
        Tmin, Tmax: tuple
            The minimum and maximum periods for the x-axis
        """
        y_min = max(self.tf.rho.min(), 1e-20)
        y_max = max(self.tf.rho.max(), 1e-20)

        yy_min = 10 ** (np.floor(np.log10(y_min)))
        if (np.log10(y_min) - np.log10(yy_min)) < 0.15:
            yy_min = 10 ** (np.log10(yy_min) - 0.3)

        yy_max = 10 ** (np.ceil(np.log10(y_max)))
        if (np.log10(yy_max) - np.log10(y_max)) < 0.15:
            yy_max = 10 ** (np.log10(yy_max) + 0.3)

        return yy_min, yy_max

    def set_lims(self) -> list:
        """
        Set limits for the plotting axes - to be depreacted for MTpy plotting functions.

        Returns
        -------
        lims: list
            The plotting limits for period, rho and phi.
            [x_max, x_min, y_max, y_min, 0, 90]
        """
        period_min, period_max = self.set_period_limits()  # get limits for the x-axis
        rho_min, rho_max = self.set_rho_limits()
        phi_min = 0
        phi_max = 90

        if abs(rho_max - rho_min) <= 1:
            rho_min = 0.01
            rho_max = 1e4
        lims = [period_min, period_max, rho_min, rho_max, phi_min, phi_max]

        return lims

    def plot(self, station_id="Transfer Function", out_filename=None, **kwargs):
        """
        Plot the apparent resistivity and phase.

        Parameters
        ----------
        station_id: str

        Returns
        -------
        fig: matplotlib.figure.Figure
            The figure object containing the plots
        """
        fig, axs = plt.subplots(nrows=2)
        fig.suptitle(f"Station: {station_id}", fontsize=16, fontweight="demi")

        ax_res = self.rho_sub_plot(axs[0], ttl_str="", pred=None)
        ax_phase = self.phase_sub_plot(axs[1], ttl_str="", pred=None)

        for ax in [ax_res, ax_phase]:
            ax.grid(
                which="both", linestyle="--", linewidth=0.5, color="gray", alpha=0.7
            )
        plt.tight_layout()
        plt.show()

        if out_filename is not None:
            fig.savefig(out_filename, **kwargs)
        return fig



"""
    In this module are some helper functions for working with FDSN inventory objects.
"""

from loguru import logger
from typing import Optional
import obspy


def describe_inventory_stages(
    inventory: obspy.Inventory,
    assign_names: Optional[bool] = False,
    verbose: Optional[bool] = False,
) -> None:

    """
    Scans inventory looking for stages.  Has option to assign names to stages,
    these names are used as keys in MTH5. Modifies inventory in place.

    Parameters
    ----------
    inventory: obspy.Inventory
        obspy representation of station XML
    assign_names: bool
        If True, inventory stage names will be assigned
    verbose: bool
        If True, some info logs are created.

    """
    new_names_were_assigned = False
    networks = inventory.networks
    for network in networks:
        for station in network:
            for channel in station:
                response = channel.response
                stages = response.response_stages
                if verbose:
                    info = (
                        f"{network.code}-{station.code}-{channel.code}"
                        f" {len(stages)}-stage response"
                    )
                    logger.info(info)

                for i, stage in enumerate(stages):
                    if verbose:
                        logger.info(f"stagename {stage.name}")
                    if stage.name is None:
                        if assign_names:
                            new_names_were_assigned = True
                            new_name = f"{station.code}_{channel.code}_{i}"
                            stage.name = new_name
                            if verbose:
                                logger.info(
                                    f"ASSIGNING stage {stage}, name {stage.name}"
                                )
                    if hasattr(stage, "symmetry"):
                        pass
                        # import matplotlib.pyplot as plt
                        # print(f"symmetry: {stage.symmetry}")
                        # plt.figure()
                        # plt.clf()
                        # plt.plot(stage.coefficients)
                        # plt.ylabel("Filter Amplitude")
                        # plt.xlabel("Filter 'Tap'")
                        # plt.title(f"{stage.name}; symmetry: {stage.symmetry}")
                        # plt.savefig(f"{stage.name}.png"))
                        # plt.show()
    if new_names_were_assigned:
        inventory.networks = networks
        logger.info("Inventory Networks Reassigned")
    return


def scan_inventory_for_nonconformity(inventory, verbose=False):
    """
    One off method for dealing with issues of historical data.
    Checks for the following:
    1. Channel Codes: Q2, Q3 --> Q1, Q2
    2. Field-type code: "T" instead of "F"
    3. Tesla to nT
    Parameters
    ----------
    inventory : obspy.core.inventory.inventory.Inventory
        Object containing metadata about station and channels

    Returns
    -------
    inventory : obspy.core.inventory.inventory.Inventory
        Object containing metadata about station and channels
        Might be modified during this function
    """
    networks = inventory.networks
    for network in networks:
        for station in network:
            channel_codes = [x.code[1:3] for x in station.channels]
            if verbose:
                logger.info(channel_codes)

            # Electric channel remap {Q2, Q3}-->{Q1, Q2}>
            if ("Q2" in channel_codes) & ("Q3" in channel_codes):
                if verbose:
                    logger.info(
                        "Detected a likely non-FDSN conformant convnetion "
                        "unless there is a vertical electric dipole"
                    )
                    logger.info("Fixing Electric channel codes")
                # run the loop twice so don't accidentally map Q3 to Q2 and Q2 to Q3
                for channel in station.channels:
                    if channel.code[1:3] == "Q2":
                        channel._code = f"{channel.code[0]}Q1"
                for channel in station.channels:
                    if channel.code[1:3] == "Q3":
                        channel._code = f"{channel.code[0]}Q2"
                # print("Applied unstable fix to electric channel names")
                logger.info("{Q2, Q3} --> {Q1, Q2}")

            # Magnetic channle remap {T1,T2,T3}-->{F1, F2, F3}
            cond1 = "T1" in channel_codes
            cond2 = "T2" in channel_codes
            cond3 = "T3" in channel_codes
            if cond1 or cond2 or cond3:
                if verbose:
                    logger.info("Detected a likely non-FDSN conformant convention ")
                    # unless there is Tidal data (channel code T)
                    logger.info("Fixing Magnetic channel codes")
                for channel in station.channels:
                    if channel.code[1] == "T":
                        channel._code = f"{channel.code[0]}F{channel.code[2]}"
                # print("Applied unstable fix to magnetic channel names")
                logger.info("{T1,T2,T3} --> {F1, F2, F3}")

            # Below represents correct syntax for a valid, but not conventional way to correct units
            # leaving the code (commented) for future reference
            # # Tesla to nanoTesla
            # for channel in station:
            #     response = channel.response
            #     for stage in response.response_stages:
            #         if verbose:
            #             msg = f"{channel.code} {stage.stage_sequence_number}"
            #             msg = f"{msg} {stage.input_units}"
            #             logger.info(msg)
            #         if stage.input_units == "T":
            #             stage.input_units = "nT"
            #             stage.stage_gain *= 1e-9
    return inventory



"""
    This module contains methods for building an MTH5 file from data at Parkfield (PKD) and Hollister
    (SAO) long term monitoring stations to use as test data.

"""
import pathlib

from aurora.test_utils.dataset_definitions import TEST_DATA_SET_CONFIGS
from aurora.sandbox.io_helpers.make_mth5_helpers import create_from_server_multistation
from aurora.test_utils.parkfield.path_helpers import PARKFIELD_PATHS
from loguru import logger
from mth5.utils.helpers import read_back_data
from mth5.helpers import close_open_files
from typing import Optional, Union

DATA_SOURCES = ["NCEDC", "https://service.ncedc.org/"]
DATASET_ID = "pkd_sao_test_00"
FDSN_DATASET = TEST_DATA_SET_CONFIGS[DATASET_ID]


def select_data_source() -> None:
    """
    Identifies appropriate web client to use for NCEDC data requests.

    This was used for debugging data access issues in the past -- may no longer be needed.

    Returns
    -------
    data_source: str
        A responsive NCEDC client.
    """
    from obspy.clients.fdsn import Client

    ok = False
    while not ok:
        for data_source in DATA_SOURCES:
            try:
                Client(base_url=data_source, force_redirect=True)
                ok = True
            except:
                logger.warning(f"Data source {data_source} not initializing")
    if not ok:
        msg = "No data sources for Parkfield / Hollister initializing\n"
        msg += "NCEDC probably down"
        logger.error(msg)
        raise ValueError(msg)
    else:
        return data_source


def make_pkdsao_mth5(
    fdsn_dataset: FDSN_DATASET,
    target_folder: Optional[Union[str, pathlib.Path]] = PARKFIELD_PATHS["data"],
) -> pathlib.Path:
    """
    Makes MTH5 file with data from Parkfield and Hollister stations to use for testing.
    """
    close_open_files()
    fdsn_dataset.data_source = select_data_source()
    fdsn_dataset.initialize_client()
    h5_path = create_from_server_multistation(
        fdsn_dataset,
        target_folder=target_folder,
        triage_units=["V/m to mV/km", "T to nT"],
    )

    for station in fdsn_dataset.station.split(","):
        logger.info(station)
        read_back_data(h5_path, station, "001")
    return h5_path


def ensure_h5_exists(
    target_folder: Union[str, pathlib.Path, None] = PARKFIELD_PATHS["data"],
) -> pathlib.Path:
    """

    Make sure that the PKD SAO MTH5 file exists.  If it does not, build it.

    Parameters
    ----------
    h5_path: Union[pathlib.Path, None]
        The target path to build to mth5.

    Returns
    -------
    h5_path: pathlib.Path
        The path to the PKD SAO mth5 file to be used for testing.

    """
    h5_path = target_folder.joinpath(FDSN_DATASET.h5_filebase)
    if h5_path.exists():
        msg = f"h5 file {h5_path} already exists -- skipping make"
        logger.info(msg)
        return h5_path

    try:
        h5_path = make_pkdsao_mth5(FDSN_DATASET, target_folder=target_folder)
        return h5_path
    except Exception as e:
        logger.error(f"Encountered {e} Exception - make_pkdsao_mth5 failed")
        logger.error("Check data server connection")
        raise IOError


def main():
    """allows the make to be run by calling this module from the command line"""
    make_pkdsao_mth5(FDSN_DATASET)


if __name__ == "__main__":
    main()



"""
This module contains the main methods used in processing mth5 objects to transfer functions.

The main function is called process_mth5.
This function was recently changed to process_mth5_legacy, os that process_mth5
can be repurposed for other TF estimation schemes.  The "legacy" version
corresponds to aurora default processing.

Notes on process_mth5_legacy:
Note 1: process_mth5 assumes application of cascading decimation, and that the
decimated data will be accessed from the previous decimation level.  This should be
revisited. It may make more sense to have a get_decimation_level() interface that
provides an option of applying decimation or loading pre-decimated data.
This will be addressed via creation of the FC layer inside mth5.

Note 2: We can encounter cases where some runs can be decimated and others can not.
We need a way to handle this. For example, a short run may not yield any data from a
later decimation level. An attempt to handle this has been made in TF Kernel by
adding a is_valid_dataset column, associated with each run-decimation level pair.

Note 3: This point in the loop marks the interface between _generation_ of the FCs and
 their _usage_. In future the code above this comment would be pushed into
 the creation of the spectrograms and the code below this would access those FCs and
 execute compute_transfer_function().
 This would also be an appropriate place to place a feature extraction layer, and
 compute weights for the FCs.

"""

from typing import Optional, Tuple, Union

import xarray as xr
from loguru import logger
from mth5.helpers import close_open_files

import aurora.config.metadata.processing

# =============================================================================
# Imports
# =============================================================================
from aurora.pipelines.feature_weights import calculate_weights, extract_features
from aurora.pipelines.transfer_function_helpers import (
    process_transfer_functions,
    process_transfer_functions_with_weights,
)
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.time_series.spectrogram_helpers import get_spectrograms, merge_stfts
from aurora.transfer_function.transfer_function_collection import (
    TransferFunctionCollection,
)
from aurora.transfer_function.TTFZ import TTFZ


SUPPORTED_PROCESSINGS = [
    "legacy",
]

# =============================================================================


def process_tf_decimation_level(
    config: aurora.config.metadata.processing.Processing,
    i_dec_level: int,
    local_stft_obj: xr.core.dataset.Dataset,
    remote_stft_obj: Union[xr.core.dataset.Dataset, None],
    weights: Optional[
        Tuple[str]
    ] = None,  # TODO: Make this a Literal of SupportedWeights
    units="MT",
):
    """
    Processing pipeline for a single decimation_level

    TODO: Add a check that the processing config sample rates agree with the data
    sampling rates otherwise raise Exception
    TODO: Add units to local_stft_obj, remote_stft_obj
    TODO: This is the method that should be accessing weights
    This method can be single station or remote based on the process cfg

    Parameters
    ----------
    config: aurora.config.metadata.processing.Processing,
        Config for a single decimation level
    i_dec_level: int
        decimation level_id
        ?could we pack this into the decimation level as an attr?
    local_stft_obj: xarray.core.dataset.Dataset
        The time series of Fourier coefficients from the local station
    remote_stft_obj: xarray.core.dataset.Dataset or None
        The time series of Fourier coefficients from the remote station
    units: str
        one of ["MT","SI"]

    Returns
    -------
    transfer_function_obj : aurora.transfer_function.TTFZ.TTFZ
        The transfer function values packed into an object
    """
    frequency_bands = config.decimations[i_dec_level].frequency_bands_obj()
    transfer_function_obj = TTFZ(i_dec_level, frequency_bands, processing_config=config)
    dec_level_config = config.decimations[i_dec_level]

    try:
        transfer_function_obj = process_transfer_functions_with_weights(
            dec_level_config=dec_level_config,
            local_stft_obj=local_stft_obj,
            remote_stft_obj=remote_stft_obj,
            transfer_function_obj=transfer_function_obj,
        )
    except Exception as e:
        msg = (
            f"Processing transfer functions with weights failed for decimation level {i_dec_level} "
            f"with exception: {e}"
        )
        logger.warning(msg)
        try:
            transfer_function_obj = process_transfer_functions(
                dec_level_config=dec_level_config,
                local_stft_obj=local_stft_obj,
                remote_stft_obj=remote_stft_obj,
                transfer_function_obj=transfer_function_obj,
            )
        except Exception as e:
            msg = (
                f"Processing transfer functions without weights also failed for decimation level {i_dec_level} "
                f"with exception: {e}"
            )
            logger.error(msg)
            logger.exception(msg)
            raise e
    return transfer_function_obj


def process_mth5_legacy(
    config,
    tfk_dataset=None,
    units="MT",
    show_plot=False,
    z_file_path=None,
    return_collection=False,
):
    """
    This is the main method used to transform a processing_config,
    and a kernel_dataset into a transfer function estimate.

    Parameters
    ----------
    config: mt_metadata.processing.aurora.Processing or path to json
        All processing parameters
    tfk_dataset: aurora.tf_kernel.dataset.Dataset or None
        Specifies what datasets to process according to config
    units: string
        "MT" or "SI".  To be deprecated once data have units embedded
    show_plot: boolean
        Only used for dev
    z_file_path: string or pathlib.Path
        Target path for a z_file output if desired
    return_collection : boolean
        return_collection=False will return an mt_metadata TF object
        return_collection=True will return
        aurora.transfer_function.transfer_function_collection.TransferFunctionCollection

    Returns
    -------
    tf_collection: TransferFunctionCollection or mt_metadata TF
        The transfer function object
    tf_cls: mt_metadata.transfer_functions.TF
        TF object
    """
    # Initialize config and mth5s
    tfk = TransferFunctionKernel(dataset=tfk_dataset, config=config)
    tfk.update_processing_summary()
    tfk.show_processing_summary()
    tfk.validate()

    tfk.initialize_mth5s()

    msg = (
        f"Processing config indicates {len(tfk.config.decimations)} "
        f"decimation levels"
    )
    logger.info(msg)
    tf_dict = {}

    for i_dec_level, dec_level_config in enumerate(tfk.valid_decimations()):
        # if not tfk.all_fcs_already_exist():
        tfk.update_dataset_df(i_dec_level)  # TODO: could add noise here
        tfk.apply_clock_zero(dec_level_config)

        stfts = get_spectrograms(tfk, i_dec_level, units=units)

        local_merged_stft_obj, remote_merged_stft_obj = merge_stfts(stfts, tfk)

        # FC TF Interface here (see Note #3)
        try:
            # Feature Extraction, Selection of weights
            extract_features(dec_level_config, tfk_dataset)
            calculate_weights(dec_level_config, tfk_dataset)
        except Exception as e:
            msg = f"Feature weights calculation Failed -- procesing without weights -- {e}"
            # logger.warning(msg)
            logger.exception(msg)
        try:
            ttfz_obj = process_tf_decimation_level(
                tfk.config,
                i_dec_level,
                local_merged_stft_obj,
                remote_merged_stft_obj,
            )
        except Exception as e:
            msg = (
                f"Processing transfer functions failed for decimation level {i_dec_level} "
                f"with exception: {e}. Skipping this decimation level."
            )
            logger.error(msg)
            logger.exception(msg)
            continue
        ttfz_obj.apparent_resistivity(tfk.config.channel_nomenclature, units=units)
        tf_dict[i_dec_level] = ttfz_obj

        if show_plot:
            fig = ttfz_obj.plot()

    tf_collection = TransferFunctionCollection(
        tf_dict=tf_dict, processing_config=tfk.config
    )

    try:
        tf_cls = tfk.export_tf_collection(tf_collection)
        logger.info(f"type(tf_cls): {type(tf_cls)}")
        if z_file_path:
            tf_cls.write(z_file_path)
            logger.info(f"Transfer function object written to {z_file_path}")
    except Exception as e:
        msg = "TF collection could not export to mt_metadata TransferFunction\n"
        msg += f"Failed with exception {e}\n"
        msg += "Perhaps an unconventional mixture of input/output channels was used\n"
        msg += f"Input channels were {tfk.config.decimations[0].input_channels}\n"
        msg += f"Output channels were {tfk.config.decimations[0].output_channels}\n"
        msg += "No z-file will be written in this case\n"
        msg += "Will return a legacy TransferFunctionCollection object, not mt_metadata object."
        logger.error(msg)
        return_collection = True

    tfk.dataset.close_mth5s()
    if return_collection:
        return tf_collection  # mostly used for debugging and may be deprecated.
    else:
        return tf_cls


def process_mth5(
    config,
    tfk_dataset=None,
    units="MT",
    show_plot=False,
    z_file_path=None,
    return_collection=False,
    processing_type="legacy",
):
    """
    This is a pass-through method that routes the config and tfk_dataset to MT data processing.
    It currently only supports legacy aurora processing.

    Parameters
    ----------
    config: mt_metadata.processing.aurora.Processing or path to json
        All processing parameters
    tfk_dataset: aurora.tf_kernel.dataset.Dataset or None
        Specifies what datasets to process according to config
    units: string
        "MT" or "SI".  To be deprecated once data have units embedded
    show_plot: boolean
        Only used for dev
    z_file_path: string or pathlib.Path
        Target path for a z_file output if desired
    return_collection : boolean
        return_collection=False will return an mt_metadata TF object
        return_collection=True will return
        aurora.transfer_function.transfer_function_collection.TransferFunctionCollection
    processing_type: string
        Controlled vocabulary, must be one of ["legacy",]
        This is not really supported now, but the idea is that in future, the config and tfk_dataset can be passed to
        another processing method if desired.

    Returns
    -------
    tf_obj: TransferFunctionCollection or mt_metadata.transfer_functions.TF
        The transfer function object
    """
    if processing_type not in SUPPORTED_PROCESSINGS:
        raise NotImplementedError(f"Processing type {processing_type} not supported")

    if processing_type == "legacy":
        try:
            return process_mth5_legacy(
                config,
                tfk_dataset=tfk_dataset,
                units=units,
                show_plot=show_plot,
                z_file_path=z_file_path,
                return_collection=return_collection,
            )
        except Exception as e:
            close_open_files()
            msg = "Failed to run legacy processing\n"
            msg += "closing all open mth5 files and exiting"
            msg += f"The encountered exception was {e}"
            logger.error(msg)
            logger.exception(msg)
            return



"""
Tests for complete Aurora processing workflow using CAS04 data.

Tests the pipeline:
1. MTH5 file → RunSummary → KernelDataset
2. ConfigCreator → processing config
3. process_mth5() → TransferFunction
4. Compare results to EMTF reference

This extends the testing from test_processing_workflow_cas04.py with actual processing
and comparison to EMTF results.
"""

from pathlib import Path

import numpy as np
import pytest
from mt_metadata.transfer_functions.core import TF
from mth5.processing import KernelDataset, RunSummary

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.transfer_function.compare import CompareTF


# ============================================================================
# Helper Functions
# ============================================================================


def _validate_emtf_comparison(
    comparison_result,
    subtests,
    z_ratio=(0.8, 1.2),
    z_std_limit=1.5,
    t_ratio=(0.8, 1.6),
    t_std_limit=0.5,
):
    """Helper function to validate transfer function comparison results.

    Args:
        comparison_result: Result dictionary from compare_transfer_functions()
        subtests: pytest subtests fixture
        z_ratio: Tuple of (min, max) acceptable impedance magnitude ratios
        z_std_limit: Maximum acceptable impedance standard deviation
        t_ratio: Tuple of (min, max) acceptable tipper magnitude ratios
        t_std_limit: Maximum acceptable tipper standard deviation
    """
    # Check impedance if present
    if comparison_result["impedance_ratio"] is not None:
        for ii in range(2):
            for jj in range(2):
                if ii != jj:
                    key = f"Z_{ii}{jj}"
                    with subtests.test(
                        msg=f"Checking impedance magnitude ratio for {key}"
                    ):
                        assert (
                            z_ratio[0]
                            < comparison_result["impedance_ratio"][key]
                            < z_ratio[1]
                        ), f"{key} impedance magnitudes differ significantly. Median ratio: {comparison_result['impedance_ratio'][key]:.3f}"

                    with subtests.test(msg=f"Checking impedance std for {key}"):
                        assert (
                            comparison_result["impedance_std"][key] < z_std_limit
                        ), f"{key} impedance magnitudes have high standard deviation: {comparison_result['impedance_std'][key]:.3f}"

    # Check tipper if present
    if comparison_result["tipper_ratio"] is not None:
        for ii in range(1):
            for jj in range(2):
                if ii != jj:
                    key = f"T_{ii}{jj}"
                    with subtests.test(
                        msg=f"Checking tipper magnitude ratio for {key}"
                    ):
                        assert (
                            t_ratio[0]
                            < comparison_result["tipper_ratio"][key]
                            < t_ratio[1]
                        ), f"{key} tipper magnitudes differ significantly. Median ratio: {comparison_result['tipper_ratio'][key]:.3f}"

                    with subtests.test(msg=f"Checking tipper std for {key}"):
                        assert (
                            comparison_result["tipper_std"][key] < t_std_limit
                        ), f"{key} tipper magnitudes have high standard deviation: {comparison_result['tipper_std'][key]:.3f}"


# ============================================================================
# Fixtures
# ============================================================================


@pytest.fixture(scope="session")
def cas04_emtf_reference():
    """Load EMTF reference result for CAS04 - skip if validation fails."""
    emtf_file = (
        Path(__file__).parent
        / "emtf_results"
        / "CAS04-CAS04bcd_REV06-CAS04bcd_NVR08.zmm"
    )

    if not emtf_file.exists():
        pytest.skip(f"EMTF reference file not found: {emtf_file}")

    try:
        tf_emtf = TF()
        tf_emtf.read(emtf_file)
        return tf_emtf
    except Exception as e:
        pytest.skip(f"Could not read EMTF file (pydantic validation issue): {e}")


# Separate session fixtures for v010 and v020 to enable better parallelization
@pytest.fixture(scope="session")
def session_cas04_run_summary_v010(global_fdsn_miniseed_v010):
    """Session-scoped RunSummary for v0.1.0."""
    run_summary = RunSummary()
    run_summary.from_mth5s([global_fdsn_miniseed_v010])
    return run_summary


@pytest.fixture(scope="session")
def session_cas04_run_summary_v020(global_fdsn_miniseed_v020):
    """Session-scoped RunSummary for v0.2.0."""
    run_summary = RunSummary()
    run_summary.from_mth5s([global_fdsn_miniseed_v020])
    return run_summary


@pytest.fixture(scope="session")
def session_cas04_kernel_dataset_v010(session_cas04_run_summary_v010):
    """Session-scoped KernelDataset for v0.1.0."""
    kd = KernelDataset()
    kd.from_run_summary(session_cas04_run_summary_v010, "CAS04")
    return kd


@pytest.fixture(scope="session")
def session_cas04_kernel_dataset_v020(session_cas04_run_summary_v020):
    """Session-scoped KernelDataset for v0.2.0."""
    kd = KernelDataset()
    kd.from_run_summary(session_cas04_run_summary_v020, "CAS04")
    return kd


@pytest.fixture(scope="session")
def session_cas04_config_v010(session_cas04_kernel_dataset_v010):
    """Session-scoped processing config for v0.1.0."""
    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(session_cas04_kernel_dataset_v010)
    return config


@pytest.fixture(scope="session")
def session_cas04_config_v020(session_cas04_kernel_dataset_v020):
    """Session-scoped processing config for v0.2.0."""
    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(session_cas04_kernel_dataset_v020)
    return config


@pytest.fixture(scope="session")
def session_cas04_tf_result_v010(
    session_cas04_kernel_dataset_v010, session_cas04_config_v010, tmp_path_factory
):
    """Session-scoped processed TF result for v0.1.0."""
    temp_dir = tmp_path_factory.mktemp("cas04_processing_v010")
    z_file_path = temp_dir / "CAS04_v010.zss"

    tf_result = process_mth5(
        session_cas04_config_v010,
        session_cas04_kernel_dataset_v010,
        units="MT",
        show_plot=False,
        z_file_path=z_file_path,
    )
    return tf_result


@pytest.fixture(scope="session")
def session_cas04_tf_result_v020(
    session_cas04_kernel_dataset_v020, session_cas04_config_v020, tmp_path_factory
):
    """Session-scoped processed TF result for v0.2.0."""
    temp_dir = tmp_path_factory.mktemp("cas04_processing_v020")
    z_file_path = temp_dir / "CAS04_v020.zss"

    tf_result = process_mth5(
        session_cas04_config_v020,
        session_cas04_kernel_dataset_v020,
        units="MT",
        show_plot=False,
        z_file_path=z_file_path,
    )
    return tf_result


# Selector fixtures that choose based on version parameter
@pytest.fixture
def cas04_run_summary(request):
    """Select appropriate RunSummary based on version."""
    version = request.param if hasattr(request, "param") else "v010"
    if version == "v010":
        fixture = request.getfixturevalue("session_cas04_run_summary_v010")
    else:
        fixture = request.getfixturevalue("session_cas04_run_summary_v020")
    return fixture.clone()


@pytest.fixture
def cas04_kernel_dataset(request):
    """Select appropriate KernelDataset based on version."""
    version = request.param if hasattr(request, "param") else "v010"
    if version == "v010":
        return request.getfixturevalue("session_cas04_kernel_dataset_v010")
    else:
        return request.getfixturevalue("session_cas04_kernel_dataset_v020")


@pytest.fixture
def cas04_config(request):
    """Select appropriate config based on version."""
    version = request.param if hasattr(request, "param") else "v010"
    if version == "v010":
        return request.getfixturevalue("session_cas04_config_v010")
    else:
        return request.getfixturevalue("session_cas04_config_v020")


@pytest.fixture
def session_cas04_tf_result(request):
    """Select appropriate TF result based on version."""
    version = request.param if hasattr(request, "param") else "v010"
    if version == "v010":
        return request.getfixturevalue("session_cas04_tf_result_v010")
    else:
        return request.getfixturevalue("session_cas04_tf_result_v020")


@pytest.fixture
def temp_output_dir(tmp_path):
    """Temporary directory for output files."""
    return tmp_path


@pytest.fixture(scope="session")
def session_interpolated_comparison_v010(
    session_cas04_tf_result_v010, cas04_emtf_reference
):
    """Session-scoped interpolated TF comparison for v0.1.0."""
    if cas04_emtf_reference is None:
        pytest.skip("EMTF reference not available")
    return CompareTF(session_cas04_tf_result_v010, cas04_emtf_reference)


@pytest.fixture(scope="session")
def session_interpolated_comparison_v020(
    session_cas04_tf_result_v020, cas04_emtf_reference
):
    """Session-scoped interpolated TF comparison for v0.2.0."""
    if cas04_emtf_reference is None:
        pytest.skip("EMTF reference not available")
    return CompareTF(session_cas04_tf_result_v020, cas04_emtf_reference)


@pytest.fixture
def session_interpolated_comparison(request):
    """Select appropriate interpolated comparison based on version."""
    version = request.param if hasattr(request, "param") else "v010"
    if version == "v010":
        return request.getfixturevalue("session_interpolated_comparison_v010")
    else:
        return request.getfixturevalue("session_interpolated_comparison_v020")


# Test Classes


@pytest.mark.parametrize("cas04_config", ["v010", "v020"], indirect=True)
class TestConfigCreation:
    """Test configuration creation from KernelDataset."""

    def test_config_creator_from_kernel_dataset(self, cas04_config):
        """Test ConfigCreator can create config from KernelDataset."""
        assert cas04_config is not None
        assert hasattr(cas04_config, "decimations")
        assert len(cas04_config.decimations) > 0

    def test_config_has_required_attributes(self, cas04_config):
        """Test that created config has all required attributes."""
        # Config should have key attributes
        assert hasattr(cas04_config, "decimations")
        assert hasattr(cas04_config, "stations")
        assert len(cas04_config.stations) > 0

    def test_config_decimation_levels(self, cas04_config):
        """Test config has reasonable decimation levels."""
        # Should have at least one decimation level
        assert len(cas04_config.decimations) > 0

        # Each decimation should have bands defined
        for dec in cas04_config.decimations:
            # Decimations should have frequency bands
            assert hasattr(dec, "bands") or "bands" in str(dec)

    def test_can_create_processing_components(self, cas04_kernel_dataset, cas04_config):
        """Test that all processing components can be created."""
        assert cas04_config is not None
        assert cas04_kernel_dataset is not None
        assert cas04_kernel_dataset.df is not None


@pytest.mark.parametrize("session_cas04_tf_result", ["v010", "v020"], indirect=True)
class TestProcessingWorkflow:
    """Test the complete processing workflow using process_mth5."""

    def test_process_mth5_runs_successfully(self, session_cas04_tf_result):
        """Test that process_mth5 runs without errors."""
        assert session_cas04_tf_result is not None

    def test_process_mth5_returns_tf_object(self, session_cas04_tf_result):
        """Test that process_mth5 returns proper TF object."""
        assert isinstance(session_cas04_tf_result, TF)

    def test_tf_has_impedance_data(self, session_cas04_tf_result):
        """Test that resulting TF has impedance data."""
        # Check impedance exists and has correct shape
        assert hasattr(session_cas04_tf_result, "impedance")
        assert session_cas04_tf_result.impedance is not None
        assert len(session_cas04_tf_result.period) > 0

    def test_tf_has_valid_frequencies(self, session_cas04_tf_result):
        """Test that TF has valid frequency values."""
        # Check frequencies are positive and monotonic
        periods = session_cas04_tf_result.period
        assert len(periods) > 0
        assert np.all(periods > 0)

    def test_tf_channel_metadata(self, session_cas04_tf_result, subtests):
        """Test that expected channels are present in TF."""
        expected_channels = ["ex", "ey", "hx", "hy", "hz"]
        invalid_time = "1980-01-01T00:00:00"

        for chan in expected_channels:
            ch_metadata = session_cas04_tf_result.run_metadata.channels[chan]
            with subtests.test(msg=f"Checking channel metadata for {chan}"):
                assert (
                    ch_metadata.time_period.start != invalid_time
                ), f"Channel {chan} has invalid start time"
                assert (
                    ch_metadata.time_period.end != invalid_time
                ), f"Channel {chan} has invalid end time"
                assert (
                    ch_metadata.sample_rate > 0
                ), f"Channel {chan} sample rate should be positive"


class TestEMTFComparison:
    """Test comparison with EMTF reference results."""

    def test_emtf_reference_loads(self, cas04_emtf_reference):
        """Test that EMTF reference file can be loaded."""
        assert cas04_emtf_reference is not None
        assert hasattr(cas04_emtf_reference, "impedance")

    @pytest.mark.parametrize(
        "session_interpolated_comparison", ["v010", "v020"], indirect=True
    )
    def test_comparison(self, session_interpolated_comparison, subtests):
        """Test that impedance magnitudes are comparable between Aurora and EMTF."""
        # Use pre-computed interpolated data from session fixture
        result = session_interpolated_comparison.compare_transfer_functions()

        # Validate comparison using helper function
        _validate_emtf_comparison(
            result,
            subtests,
            z_ratio=(0.8, 1.2),
            z_std_limit=1.5,
            t_ratio=(0.8, 1.6),
            t_std_limit=0.5,
        )


@pytest.mark.parametrize("session_cas04_tf_result", ["v010", "v020"], indirect=True)
class TestDataQuality:
    """Test data quality metrics from processing."""

    def test_tf_has_error_estimates(self, session_cas04_tf_result):
        """Test that TF includes error estimates."""
        # Check for error estimates
        assert hasattr(session_cas04_tf_result, "impedance_error")
        assert session_cas04_tf_result.impedance_error is not None

    def test_errors_are_positive(self, session_cas04_tf_result):
        """Test that error estimates are positive."""
        errors = session_cas04_tf_result.impedance_error
        error_values = errors.values if hasattr(errors, "values") else errors
        assert np.all(error_values[~np.isnan(error_values)] >= 0)


class TestEndToEndIntegration:
    """End-to-end integration tests."""

    @pytest.mark.slow
    @pytest.mark.parametrize("cas04_run_summary", ["v010", "v020"], indirect=True)
    def test_complete_pipeline_from_run_summary(
        self, cas04_run_summary, temp_output_dir
    ):
        """
        Test complete pipeline from RunSummary to TF.

        Note: This test validates the integration path independently from
        session fixtures. Marked 'slow' (~40s per version).
        Run with: pytest -m slow | Skip with: pytest -m "not slow"
        """
        # Create KernelDataset
        kd = KernelDataset()
        kd.from_run_summary(cas04_run_summary, "CAS04")

        # Create config
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(kd)

        # Process
        z_file_path = temp_output_dir / "CAS04_integration.zss"

        tf_result = process_mth5(
            config,
            kd,
            units="MT",
            show_plot=False,
            z_file_path=z_file_path,
        )

        # Verify complete result
        assert tf_result is not None
        assert len(tf_result.period) > 0
        assert z_file_path.exists()

    @pytest.mark.parametrize("session_cas04_tf_result", ["v010", "v020"], indirect=True)
    def test_can_read_written_file(self, session_cas04_tf_result, temp_output_dir):
        """Test that written z-file can be read back."""
        # Write to new file
        z_file_path = temp_output_dir / "CAS04_readback.zss"
        session_cas04_tf_result.write(z_file_path)

        # Read back
        tf_readback = TF()
        tf_readback.read(z_file_path)

        # Compare
        assert len(tf_readback.period) == len(session_cas04_tf_result.period)
        # Use decimal=5 since periods have slight floating point differences
        np.testing.assert_array_almost_equal(
            tf_readback.period, session_cas04_tf_result.period, decimal=5
        )


class TestEdgeCases:
    """Test edge cases and error handling."""

    @pytest.mark.parametrize("cas04_run_summary", ["v010", "v020"], indirect=True)
    def test_invalid_station_id_handling(self, cas04_run_summary):
        """Test handling of invalid station IDs."""
        # This should work even if station IDs don't match expected patterns
        kd = KernelDataset()
        kd.from_run_summary(cas04_run_summary, "CAS04")

        assert kd is not None
        assert kd.df is not None

    @pytest.mark.parametrize("cas04_kernel_dataset", ["v010", "v020"], indirect=True)
    def test_missing_channels_handling(self, cas04_kernel_dataset):
        """Test that processing handles missing channels gracefully."""
        # Even with limited channels, config creation should work
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(cas04_kernel_dataset)

        assert config is not None
        assert len(config.stations) > 0



# -*- coding: utf-8 -*-
"""
Pytest suite for RegressionEstimator base class.

Tests transfer function regression using fixtures and subtests.
Optimized for pytest-xdist parallel execution.
"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from aurora.transfer_function.regression.base import RegressionEstimator
from aurora.transfer_function.regression.iter_control import IterControl


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture(scope="module")
def expected_solution():
    """Expected solution for mini dataset regression."""
    return np.array([-0.04192569 - 0.36502722j, -3.65284496 - 4.05194938j])


@pytest.fixture(scope="module")
def mini_dataset_full():
    """Create full mini dataset with 3 rows."""
    ex_data = np.array(
        [
            4.39080123e-07 - 2.41097397e-06j,
            -2.33418464e-06 + 2.10752581e-06j,
            1.38642624e-06 - 1.87333571e-06j,
        ]
    )
    hx_data = np.array(
        [
            7.00767250e-07 - 9.18819198e-07j,
            -1.06648904e-07 + 8.19420154e-07j,
            -1.02700963e-07 - 3.73904463e-07j,
        ]
    )
    hy_data = np.array(
        [
            1.94321684e-07 + 3.71934877e-07j,
            1.15361101e-08 - 6.32581646e-07j,
            3.86095787e-08 + 4.33155345e-07j,
        ]
    )
    timestamps = pd.date_range(
        start=pd.Timestamp("1977-03-02T06:00:00"), periods=len(ex_data), freq="s"
    )
    frequency = 0.666 * np.ones(len(ex_data))

    df = pd.DataFrame(
        data={
            "time": timestamps,
            "frequency": frequency,
            "ex": ex_data,
            "hx": hx_data,
            "hy": hy_data,
        }
    )
    df = df.set_index(["time", "frequency"])
    return df.to_xarray()


@pytest.fixture(scope="module")
def mini_dataset_single():
    """Create mini dataset with 1 row (underdetermined)."""
    ex_data = np.array([4.39080123e-07 - 2.41097397e-06j])
    hx_data = np.array([7.00767250e-07 - 9.18819198e-07j])
    hy_data = np.array([1.94321684e-07 + 3.71934877e-07j])

    timestamps = pd.date_range(
        start=pd.Timestamp("1977-03-02T06:00:00"), periods=len(ex_data), freq="s"
    )
    frequency = 0.666 * np.ones(len(ex_data))

    df = pd.DataFrame(
        data={
            "time": timestamps,
            "frequency": frequency,
            "ex": ex_data,
            "hx": hx_data,
            "hy": hy_data,
        }
    )
    df = df.set_index(["time", "frequency"])
    return df.to_xarray()


@pytest.fixture
def dataset_xy_full(mini_dataset_full):
    """Prepare X and Y datasets from full mini dataset."""
    X = mini_dataset_full[["hx", "hy"]]
    X = X.stack(observation=("frequency", "time"))
    Y = mini_dataset_full[["ex"]]
    Y = Y.stack(observation=("frequency", "time"))
    return X, Y


@pytest.fixture
def dataset_xy_single(mini_dataset_single):
    """Prepare X and Y datasets from single-row mini dataset."""
    X = mini_dataset_single[["hx", "hy"]]
    X = X.stack(observation=("frequency", "time"))
    Y = mini_dataset_single[["ex"]]
    Y = Y.stack(observation=("frequency", "time"))
    return X, Y


@pytest.fixture
def regression_estimator(dataset_xy_full):
    """Create a basic RegressionEstimator instance."""
    X, Y = dataset_xy_full
    return RegressionEstimator(X=X, Y=Y)


@pytest.fixture
def simple_regression_data():
    """Create simple synthetic regression data."""
    np.random.seed(100)
    n_obs = 20
    X = np.random.randn(2, n_obs) + 1j * np.random.randn(2, n_obs)
    true_b = np.array([[1.5 + 0.5j], [-0.8 + 1.2j]])
    Y = true_b.T @ X
    return X, Y, true_b


# =============================================================================
# Test Initialization
# =============================================================================


class TestRegressionEstimatorInit:
    """Test RegressionEstimator initialization."""

    def test_init_with_xarray_dataset(self, dataset_xy_full):
        """Test initialization with xarray Dataset."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)

        assert re is not None
        assert re.X is not None
        assert re.Y is not None
        assert isinstance(re.X, np.ndarray)
        assert isinstance(re.Y, np.ndarray)

    def test_init_with_xarray_dataarray(self, dataset_xy_full):
        """Test initialization with xarray DataArray."""
        X, Y = dataset_xy_full
        X_da = X.to_array()
        Y_da = Y.to_array()

        re = RegressionEstimator(X=X_da, Y=Y_da)

        assert re is not None
        assert isinstance(re.X, np.ndarray)
        assert isinstance(re.Y, np.ndarray)

    def test_init_with_numpy_array(self, dataset_xy_full):
        """Test initialization with numpy arrays."""
        X, Y = dataset_xy_full
        X_np = X.to_array().data
        Y_np = Y.to_array().data

        re = RegressionEstimator(X=X_np, Y=Y_np)

        assert re is not None
        assert isinstance(re.X, np.ndarray)
        assert isinstance(re.Y, np.ndarray)

    def test_init_sets_attributes(self, dataset_xy_full):
        """Test that initialization sets expected attributes."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)

        assert re.b is None
        assert re.cov_nn is None
        assert re.cov_ss_inv is None
        assert re.squared_coherence is None
        assert hasattr(re, "iter_control")
        assert isinstance(re.iter_control, IterControl)

    def test_init_with_custom_iter_control(self, dataset_xy_full):
        """Test initialization with custom IterControl."""
        X, Y = dataset_xy_full
        custom_iter = IterControl(max_number_of_iterations=50)
        re = RegressionEstimator(X=X, Y=Y, iter_control=custom_iter)

        assert re.iter_control.max_number_of_iterations == 50

    def test_init_with_channel_names(self, simple_regression_data):
        """Test initialization with explicit channel names."""
        X, Y, _ = simple_regression_data
        input_names = ["hx", "hy"]
        output_names = ["ex"]

        re = RegressionEstimator(
            X=X, Y=Y, input_channel_names=input_names, output_channel_names=output_names
        )

        assert re.input_channel_names == input_names
        assert re.output_channel_names == output_names


# =============================================================================
# Test Properties
# =============================================================================


class TestRegressionEstimatorProperties:
    """Test RegressionEstimator properties."""

    def test_n_data_property(self, regression_estimator):
        """Test n_data property returns correct number of observations."""
        assert regression_estimator.n_data == 3

    def test_n_channels_in_property(self, regression_estimator):
        """Test n_channels_in property returns correct number."""
        assert regression_estimator.n_channels_in == 2

    def test_n_channels_out_property(self, regression_estimator):
        """Test n_channels_out property returns correct number."""
        assert regression_estimator.n_channels_out == 1

    def test_degrees_of_freedom_property(self, regression_estimator):
        """Test degrees_of_freedom property calculation."""
        expected_dof = regression_estimator.n_data - regression_estimator.n_channels_in
        assert regression_estimator.degrees_of_freedom == expected_dof
        assert regression_estimator.degrees_of_freedom == 1

    def test_is_underdetermined_false(self, regression_estimator):
        """Test is_underdetermined returns False for well-determined system."""
        assert regression_estimator.is_underdetermined is False

    def test_is_underdetermined_true(self, dataset_xy_single):
        """Test is_underdetermined returns True for underdetermined system."""
        X, Y = dataset_xy_single
        re = RegressionEstimator(X=X, Y=Y)
        assert re.is_underdetermined is True

    def test_input_channel_names_from_dataset(self, dataset_xy_full):
        """Test input_channel_names extracted from xarray Dataset."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        names = re.input_channel_names

        assert isinstance(names, list)
        assert len(names) == 2
        assert "hx" in names
        assert "hy" in names

    def test_output_channel_names_from_dataset(self, dataset_xy_full):
        """Test output_channel_names extracted from xarray Dataset."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        names = re.output_channel_names

        assert isinstance(names, list)
        assert len(names) == 1
        assert "ex" in names


# =============================================================================
# Test OLS Estimation
# =============================================================================


class TestOLSEstimation:
    """Test ordinary least squares estimation methods."""

    def test_estimate_ols_qr_mode(self, dataset_xy_full, expected_solution):
        """Test estimate_ols with QR mode."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols(mode="qr")

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_estimate_ols_solve_mode(self, dataset_xy_full, expected_solution):
        """Test estimate_ols with solve mode."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols(mode="solve")

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_estimate_ols_brute_force_mode(self, dataset_xy_full, expected_solution):
        """Test estimate_ols with brute_force mode."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols(mode="brute_force")

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_estimate_ols_modes_equivalent(self, dataset_xy_full, subtests):
        """Test that different OLS modes produce equivalent results."""
        X, Y = dataset_xy_full
        modes = ["qr", "solve", "brute_force"]
        results = {}

        for mode in modes:
            with subtests.test(mode=mode):
                re = RegressionEstimator(X=X, Y=Y)
                re.estimate_ols(mode=mode)
                results[mode] = re.b.copy()

        # Compare all modes to each other
        for mode1 in modes:
            for mode2 in modes:
                if mode1 != mode2:
                    assert np.allclose(results[mode1], results[mode2])

    def test_estimate_method(self, dataset_xy_full, expected_solution):
        """Test the estimate() convenience method."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate()

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_estimate_ols_returns_b(self, dataset_xy_full):
        """Test that estimate_ols returns the b matrix."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        result = re.estimate_ols()

        assert result is not None
        assert np.array_equal(result, re.b)


# =============================================================================
# Test QR Decomposition
# =============================================================================


class TestQRDecomposition:
    """Test QR decomposition functionality."""

    def test_qr_decomposition_basic(self, regression_estimator):
        """Test basic QR decomposition."""
        Q, R = regression_estimator.qr_decomposition()

        assert Q is not None
        assert R is not None
        assert isinstance(Q, np.ndarray)
        assert isinstance(R, np.ndarray)

    def test_qr_decomposition_properties(self, regression_estimator):
        """Test QR decomposition mathematical properties."""
        Q, R = regression_estimator.qr_decomposition()

        # Q should be unitary: Q^H @ Q = I
        QHQ = Q.conj().T @ Q
        assert np.allclose(QHQ, np.eye(Q.shape[1]))

        # R should be upper triangular
        assert np.allclose(R, np.triu(R))

    def test_qr_decomposition_reconstruction(self, regression_estimator):
        """Test that Q @ R reconstructs X."""
        Q, R = regression_estimator.qr_decomposition()
        X_reconstructed = Q @ R

        assert np.allclose(X_reconstructed, regression_estimator.X)

    def test_qr_decomposition_sanity_check(self, regression_estimator):
        """Test QR decomposition with sanity check enabled."""
        Q, R = regression_estimator.qr_decomposition(sanity_check=True)

        assert Q is not None
        assert R is not None

    def test_q_property(self, regression_estimator):
        """Test Q property accessor."""
        regression_estimator.qr_decomposition()
        Q = regression_estimator.Q

        assert Q is not None
        assert isinstance(Q, np.ndarray)

    def test_r_property(self, regression_estimator):
        """Test R property accessor."""
        regression_estimator.qr_decomposition()
        R = regression_estimator.R

        assert R is not None
        assert isinstance(R, np.ndarray)

    def test_qh_property(self, regression_estimator):
        """Test QH (conjugate transpose) property."""
        regression_estimator.qr_decomposition()
        QH = regression_estimator.QH
        Q = regression_estimator.Q

        assert np.allclose(QH, Q.conj().T)

    def test_qhy_property(self, regression_estimator):
        """Test QHY property."""
        regression_estimator.qr_decomposition()
        QHY = regression_estimator.QHY

        expected = regression_estimator.QH @ regression_estimator.Y
        assert np.allclose(QHY, expected)


# =============================================================================
# Test Underdetermined Systems
# =============================================================================


class TestUnderdeterminedSystems:
    """Test handling of underdetermined regression problems."""

    def test_solve_underdetermined(self, dataset_xy_single):
        """Test solve_underdetermined method."""
        X, Y = dataset_xy_single
        re = RegressionEstimator(X=X, Y=Y)
        re.solve_underdetermined()

        assert re.b is not None
        assert isinstance(re.b, np.ndarray)

    def test_underdetermined_sets_covariances(self, dataset_xy_single):
        """Test that solve_underdetermined sets covariance matrices."""
        X, Y = dataset_xy_single
        re = RegressionEstimator(X=X, Y=Y)
        # Enable return_covariance in iter_control
        re.iter_control.return_covariance = True
        re.solve_underdetermined()

        assert re.cov_nn is not None
        assert re.cov_ss_inv is not None

    def test_underdetermined_covariance_shapes(self, dataset_xy_single):
        """Test covariance matrix shapes for underdetermined system."""
        X, Y = dataset_xy_single
        re = RegressionEstimator(X=X, Y=Y)
        # Enable return_covariance in iter_control
        re.iter_control.return_covariance = True
        re.solve_underdetermined()

        assert re.cov_nn.shape == (re.n_channels_out, re.n_channels_out)
        assert re.cov_ss_inv.shape == (re.n_channels_in, re.n_channels_in)


# =============================================================================
# Test Different Input Types
# =============================================================================


class TestDifferentInputTypes:
    """Test RegressionEstimator with different input data types."""

    def test_xarray_dataset_input(self, dataset_xy_full, expected_solution):
        """Test regression with xarray Dataset input."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_xarray_dataarray_input(self, dataset_xy_full, expected_solution):
        """Test regression with xarray DataArray input."""
        X, Y = dataset_xy_full
        X_da = X.to_array()
        Y_da = Y.to_array()

        re = RegressionEstimator(X=X_da, Y=Y_da)
        re.estimate_ols()

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_numpy_array_input(self, dataset_xy_full, expected_solution):
        """Test regression with numpy array input."""
        X, Y = dataset_xy_full
        X_np = X.to_array().data
        Y_np = Y.to_array().data

        re = RegressionEstimator(X=X_np, Y=Y_np)
        re.estimate_ols()

        difference = re.b - np.atleast_2d(expected_solution).T
        assert np.allclose(difference, 0)

    def test_all_input_types_equivalent(self, dataset_xy_full):
        """Test that all input types produce equivalent results."""
        X, Y = dataset_xy_full

        # Dataset
        re_ds = RegressionEstimator(X=X, Y=Y)
        re_ds.estimate_ols()

        # DataArray
        re_da = RegressionEstimator(X=X.to_array(), Y=Y.to_array())
        re_da.estimate_ols()

        # Numpy
        re_np = RegressionEstimator(X=X.to_array().data, Y=Y.to_array().data)
        re_np.estimate_ols()

        assert np.allclose(re_ds.b, re_da.b)
        assert np.allclose(re_ds.b, re_np.b)


# =============================================================================
# Test xarray Conversion
# =============================================================================


class TestXarrayConversion:
    """Test conversion of results to xarray format."""

    def test_b_to_xarray(self, dataset_xy_full):
        """Test b_to_xarray method."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        xr_result = re.b_to_xarray()

        assert isinstance(xr_result, xr.DataArray)
        assert xr_result is not None

    def test_b_to_xarray_dimensions(self, dataset_xy_full):
        """Test b_to_xarray has correct dimensions."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        xr_result = re.b_to_xarray()

        assert "output_channel" in xr_result.dims
        assert "input_channel" in xr_result.dims

    def test_b_to_xarray_coordinates(self, dataset_xy_full):
        """Test b_to_xarray has correct coordinates."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        xr_result = re.b_to_xarray()

        assert "output_channel" in xr_result.coords
        assert "input_channel" in xr_result.coords
        assert len(xr_result.coords["input_channel"]) == 2
        assert len(xr_result.coords["output_channel"]) == 1

    def test_b_to_xarray_values(self, dataset_xy_full, expected_solution):
        """Test b_to_xarray contains correct values."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        xr_result = re.b_to_xarray()

        # Compare transposed b to xarray values
        assert np.allclose(xr_result.values, re.b.T)


# =============================================================================
# Test Data Validation
# =============================================================================


class TestDataValidation:
    """Test data validation and error handling."""

    def test_mismatched_observations_raises_error(self, mini_dataset_full):
        """Test that mismatched X and Y observations raises an error."""
        X = mini_dataset_full[["hx", "hy"]]
        X = X.stack(observation=("frequency", "time"))

        # Create Y with different number of observations
        Y_short = mini_dataset_full[["ex"]].isel(time=slice(0, 2))
        Y_short = Y_short.stack(observation=("frequency", "time"))

        with pytest.raises(Exception):
            RegressionEstimator(X=X, Y=Y_short)


# =============================================================================
# Test Numerical Stability
# =============================================================================


class TestNumericalStability:
    """Test numerical stability of regression methods."""

    def test_ols_with_synthetic_data(self, simple_regression_data):
        """Test OLS with synthetic data of known solution."""
        X, Y, true_b = simple_regression_data

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert np.allclose(re.b, true_b, rtol=1e-10)

    def test_large_magnitude_values(self):
        """Test regression with large magnitude values."""
        scale = 1e10
        np.random.seed(101)
        X = np.random.randn(2, 10) * scale + 1j * np.random.randn(2, 10) * scale
        true_b = np.array([[1.0 + 0.5j], [-0.5 + 1.0j]])
        Y = true_b.T @ X

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert np.allclose(re.b, true_b, rtol=1e-6)

    def test_small_magnitude_values(self):
        """Test regression with small magnitude values."""
        scale = 1e-10
        np.random.seed(102)
        X = np.random.randn(2, 10) * scale + 1j * np.random.randn(2, 10) * scale
        true_b = np.array([[1.0 + 0.5j], [-0.5 + 1.0j]])
        Y = true_b.T @ X

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert np.allclose(re.b, true_b, rtol=1e-6)

    def test_consistency_across_random_seeds(self, subtests):
        """Test that results are consistent across different random seeds."""
        seeds = [200, 201, 202, 203, 204]
        true_b = np.array([[1.5 + 0.3j], [-0.7 + 0.9j]])

        for seed in seeds:
            with subtests.test(seed=seed):
                np.random.seed(seed)
                X = np.random.randn(2, 15) + 1j * np.random.randn(2, 15)
                Y = true_b.T @ X

                re = RegressionEstimator(X=X, Y=Y)
                re.estimate_ols()

                assert np.allclose(re.b, true_b, rtol=1e-10)


# =============================================================================
# Test Edge Cases
# =============================================================================


class TestEdgeCases:
    """Test edge cases and boundary conditions."""

    def test_minimum_observations(self):
        """Test with minimum number of observations (n = n_channels_in)."""
        # X should be (n_channels_in, n_observations) = (2, 2)
        X = np.array([[1.0 + 0j, 3.0 + 0j], [2.0 + 0j, 4.0 + 0j]])
        # Y should be (n_channels_out, n_observations) = (1, 2)
        Y = np.array([[5.0 + 1j, 6.0 + 2j]])

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert re.b is not None
        assert np.all(np.isfinite(re.b))

    def test_single_output_channel(self, dataset_xy_full):
        """Test with single output channel."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert re.n_channels_out == 1
        assert re.b.shape[0] == re.n_channels_in

    def test_real_valued_data(self):
        """Test with real-valued (not complex) data."""
        X = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        Y = np.array([[7.0, 8.0, 9.0]])

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert re.b is not None
        assert np.all(np.isfinite(re.b))


# =============================================================================
# Test Data Integrity
# =============================================================================


class TestDataIntegrity:
    """Test that regression doesn't modify input data."""

    def test_estimate_preserves_input_X(self, dataset_xy_full):
        """Test that estimation doesn't modify input X."""
        X, Y = dataset_xy_full
        X_orig = X.copy(deep=True)

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert X.equals(X_orig)

    def test_estimate_preserves_input_Y(self, dataset_xy_full):
        """Test that estimation doesn't modify input Y."""
        X, Y = dataset_xy_full
        Y_orig = Y.copy(deep=True)

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert Y.equals(Y_orig)

    def test_qr_decomposition_preserves_X(self, regression_estimator):
        """Test that QR decomposition doesn't modify X."""
        X_orig = regression_estimator.X.copy()

        regression_estimator.qr_decomposition()

        assert np.allclose(regression_estimator.X, X_orig)


# =============================================================================
# Test Deterministic Behavior
# =============================================================================


class TestDeterministicBehavior:
    """Test that methods produce deterministic results."""

    def test_estimate_ols_deterministic(self, dataset_xy_full):
        """Test that estimate_ols produces same result on repeated calls."""
        X, Y = dataset_xy_full

        results = []
        for _ in range(5):
            re = RegressionEstimator(X=X, Y=Y)
            re.estimate_ols()
            results.append(re.b.copy())

        for result in results[1:]:
            assert np.allclose(result, results[0])

    def test_qr_decomposition_deterministic(self, dataset_xy_full):
        """Test that QR decomposition is deterministic."""
        X, Y = dataset_xy_full

        re = RegressionEstimator(X=X, Y=Y)
        Q1, R1 = re.qr_decomposition()
        Q2, R2 = re.qr_decomposition(re.X)

        assert np.allclose(Q1, Q2)
        assert np.allclose(R1, R2)


# =============================================================================
# Test Mathematical Properties
# =============================================================================


class TestMathematicalProperties:
    """Test mathematical properties of regression."""

    def test_residual_minimization(self, simple_regression_data):
        """Test that OLS minimizes the residual."""
        X, Y, _ = simple_regression_data

        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        # Compute residual
        Y_pred = re.b.T @ X
        residual = Y - Y_pred

        # For exact case (no noise), residual should be near zero
        assert np.linalg.norm(residual) < 1e-10

    def test_solution_shape(self, dataset_xy_full):
        """Test that solution has correct shape."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert re.b.shape == (re.n_channels_in, re.n_channels_out)

    def test_qr_orthogonality(self, regression_estimator):
        """Test Q matrix orthogonality from QR decomposition."""
        Q, _ = regression_estimator.qr_decomposition()

        # Q should satisfy Q^H @ Q = I
        QHQ = Q.conj().T @ Q
        identity = np.eye(Q.shape[1])

        assert np.allclose(QHQ, identity, atol=1e-10)


# =============================================================================
# Test Return Values
# =============================================================================


class TestReturnValues:
    """Test characteristics of return values."""

    def test_b_is_finite(self, dataset_xy_full):
        """Test that regression solution b contains finite values."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert np.all(np.isfinite(re.b))

    def test_b_is_complex(self, dataset_xy_full):
        """Test that regression solution b is complex."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert np.iscomplexobj(re.b)

    def test_b_not_all_zero(self, dataset_xy_full):
        """Test that regression solution b is not all zeros."""
        X, Y = dataset_xy_full
        re = RegressionEstimator(X=X, Y=Y)
        re.estimate_ols()

        assert not np.allclose(re.b, 0)



"""

This module contains methods associated with legacy EMTF z-file TF format.

Development notes:
They extract info needed to setup emtf_z files.
These methods can possibly be moved under mt_metadata, or deprecated.

"""
import pathlib
from typing import Optional, Union

from loguru import logger


EMTF_CHANNEL_ORDER = ["hx", "hy", "hz", "ex", "ey"]


def get_default_orientation_block(n_ch: int = 5) -> list:
    """
    creates a text block like the part of the z-file that holds channel orientations.

    Helper function used when working with matlab structs which do not have enough
    info to make headers

    Parameters
    ----------
    n_ch: int
        number of channels at the station

    Returns
    -------
    orientation_strs: list
        List of text strings, one per channel
    """
    orientation_strs = []
    orientation_strs.append("    1     0.00     0.00 tes  Hx\n")
    orientation_strs.append("    2    90.00     0.00 tes  Hy\n")
    if n_ch == 5:
        orientation_strs.append("    3     0.00     0.00 tes  Hz\n")
    orientation_strs.append("    4     0.00     0.00 tes  Ex\n")
    orientation_strs.append("    5    90.00     0.00 tes  Ey\n")
    return orientation_strs


def clip_bands_from_z_file(
    z_path: Union[str, pathlib.Path],
    n_bands_clip: int,
    output_z_path: Optional[Union[str, pathlib.Path, None]] = None,
    n_sensors: Optional[int] = 5,
):
    """
    This function clips periods off the end of an EMTF legacy z_file.

    Development Notes:
    It can come in handy for manipulating matlab results of synthetic data.

    Parameters
    ----------
    z_path: Path or str
        path to the z_file to read in and clip periods from
    n_periods_clip: integer
        how many periods to clip from the end of the zfile
    overwrite: bool
        whether to overwrite the zfile or rename it
    n_sensors

    Returns
    -------

    """
    if not output_z_path:
        output_z_path = z_path

    if n_sensors == 5:
        n_lines_per_period = 13
    elif n_sensors == 4:
        n_lines_per_period = 11
        logger.info("WARNING n_sensors==4 NOT TESTED")

    f = open(z_path, "r")
    lines = f.readlines()
    f.close()

    for i in range(n_bands_clip):
        lines = lines[:-n_lines_per_period]
    n_bands_str = lines[5].split()[-1]
    n_bands = int(n_bands_str)
    new_n_bands = n_bands - n_bands_clip
    new_n_bands_str = str(new_n_bands)
    lines[5] = lines[5].replace(n_bands_str, new_n_bands_str)

    f = open(output_z_path, "w")
    f.writelines(lines)
    f.close()
